"""
This module contains functions for constructing and examining doppler shift in signals
"""
from typing import Tuple

import numpy as np


def time_duration(time_vector: np.ndarray) -> float:
    """
    Compute time from max and min

    :param time_vector: array of times in seconds
    :return: duration in seconds
    """
    return np.max(time_vector) - np.min(time_vector)


def time_4d_mx(time_array: np.ndarray, space_dimensions: int) -> np.array:
    """
    Input space and convert to spacetime matrix [time x XYZ]
    i.e. if time is [1, 2, 3, 4] and dimensions is 3
    result is [[1, 1, 1], [2, 2, 2], [3, 3, 3], [4, 4, 4]]

    :param time_array: array of times in seconds
    :param space_dimensions: integer number of dimensions of space
    :return: time_array arranged vertically space_dimensions number of times
    """
    return np.array([time_array, ] * space_dimensions).transpose()


def space_4d_mx(space_column_vector: np.ndarray, time_number_samples: int) -> np.array:
    """
    Input XYZ space vector and convert to spacetime matrix [time x XYZ]

    :param space_column_vector: an array representing 3-dimensional space
    :param time_number_samples: number of time samples
    :return: spacetime matrix
    """
    return np.array([space_column_vector, ] * time_number_samples)


def hadamard_dot_product_mx(x_mx: np.ndarray, y_mx: np.ndarray) -> np.ndarray:
    """
    Hadamard column-wise dot product with summation

    :param x_mx: MxN matrix x, columns represent directional vectors
    :param y_mx: MxN matrix y
    :return: sum of matrix product over columns
    """
    return np.sum(x_mx * y_mx, 1)


def range_vector_sr(x_initial_position_vector: np.array, x_final_position_vector: np.array) -> np.array:
    """
    Start to end direction, or source vs receiver

    :param x_initial_position_vector: starting position, 3-element XYZ
    :param x_final_position_vector: ending position, 3-element XYZ
    :return: receiver - source vectors
    """
    return x_final_position_vector - x_initial_position_vector


def range_matrix_sr(x_source_mx: np.ndarray, x_receiver_mx: np.ndarray) -> np.ndarray:
    """
    start to end position, source vs receiver matrix

    :param x_source_mx: starting position, 3-element XYZ columns, t rows
    :param x_receiver_mx: ending position, 3-element XYZ columns, t rows
    :return: receiver - source matrices
    """
    return x_receiver_mx - x_source_mx


def range_hadamard(r_mx: np.ndarray) -> np.ndarray:
    """
    Compute the magnitude of range and time (Square root of Hadamard dot product over columns)

    :param r_mx: positional range matrix, 3-element XYZ columns, t rows
    :return: Square root of Hadamard dot product over columns
    """
    return np.sqrt(hadamard_dot_product_mx(r_mx, r_mx))


def range_scalar(x_source_vector: np.array, x_receiver_vector: np.array) -> float:
    """
    Compute the magnitude of the range (square root of the sum of the range vector squared, element-wise)

    :param x_source_vector: starting position, 3-element XYZ
    :param x_receiver_vector: ending position, 3-element XYZ
    :return: magnitude of the range
    """
    range_vector = range_vector_sr(x_source_vector, x_receiver_vector)
    return np.sqrt(np.sum(range_vector * range_vector))


def _get_setup(
        time_array_s: np.array,
        num_space_dimensions: int,
        src_position_vector_init_xyz_m: np.array,
        src_position_vector_final_xyz_m: np.array,
        rcvr_position_vector_init_xyz_m: np.array,
        rcvr_position_vector_final_xyz_m: np.array
) -> Tuple[int, np.array, float, float]:
    """
    Apply a doppler shift on a source moving towards the receiver

    :param time_array_s: array of source times in seconds
    :param num_space_dimensions: integer number of space dimensions. Nominal = 3, should match space position vectors.
    :param src_position_vector_init_xyz_m: initial source position in space (where it initially started)
    :param src_position_vector_final_xyz_m: final source position in space
    :param rcvr_position_vector_init_xyz_m: position in space where receiver initially started
    :param rcvr_position_vector_final_xyz_m: position in space where receiver ends up
    :return: number of time samples, matrix for time, source trajectory magnitude, receiver trajectory magnitude
    """
    temp_mx_s = time_4d_mx(time_array_s, num_space_dimensions)
    source_trajectory_m = range_scalar(src_position_vector_init_xyz_m, src_position_vector_final_xyz_m)
    receiver_trajectory_m = range_scalar(rcvr_position_vector_init_xyz_m, rcvr_position_vector_final_xyz_m)

    return len(time_array_s), temp_mx_s, source_trajectory_m, receiver_trajectory_m


def _get_velocity_mps(
        speed_mps: float,
        trajectory_m: float,
        num_samples: int,
        position_vector_init_xyz_m: np.array,
        position_vector_final_xyz_m: np.array
) -> np.array:
    """
    Compute velocity (in meter per second)

    :param speed_mps: object's speed
    :param trajectory_m: magnitude of trajectory in meters
    :param num_samples: number of samples
    :param position_vector_init_xyz_m: position in space where object initially started
    :param position_vector_final_xyz_m: position in space where object ends up
    :return: velocity in XYZ + time matrix
    """
    if speed_mps > 0:
        velocity_direction = range_vector_sr(position_vector_init_xyz_m, position_vector_final_xyz_m) / trajectory_m
        velocity_mps = speed_mps * velocity_direction
    else:
        velocity_mps = np.zeros(3)
    return space_4d_mx(velocity_mps, num_samples)


def _get_final_vals(
        spacetime_matrix: np.array,
        receiver_velocity_mx_mps: np.array,
        source_velocity_mx_mps: np.array,
        time_array_s: np.array,
        rcvr_position_vector_init_xyz_m: np.array,
        src_position_vector_init_xyz_m: np.array,
        signal_speed_mps: float,
        object_speed_mps: float,
        num_dimensions: int,
        num_samples: int,
        inverse: bool = False
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Compute array of times in seconds, Magnitude of range and time, omega over omega center

    :param spacetime_matrix: spacetime matrix
    :param time_array_s: array of times
    :param receiver_velocity_mx_mps: receiver velocity in meters per second
    :param source_velocity_mx_mps: source velocity in meters per second
    :param rcvr_position_vector_init_xyz_m: position in space where receiver initially started
    :param src_position_vector_init_xyz_m: position in space where source initially started
    :param signal_speed_mps: signal speed in meters per second
    :param object_speed_mps: object speed in meters per second
    :param num_dimensions: number of dimensions
    :param num_samples: number of samples
    :param inverse: is this an inverse calculation?  Default False
    :return: array of times in seconds, Magnitude of range and time, omega over omega center
    """
    # Initial range vector, source to receiver (sr) at start of segment
    range_initial_mx_m = space_4d_mx(rcvr_position_vector_init_xyz_m - src_position_vector_init_xyz_m, num_samples)
    # Compute object values and phase
    denom = 1. / (signal_speed_mps**2 - object_speed_mps**2)
    # Dot product contracts to 1D size
    if inverse:
        temp_range_mx_m = range_initial_mx_m + receiver_velocity_mx_mps * spacetime_matrix
        term1 = (signal_speed_mps**2) * time_array_s - hadamard_dot_product_mx(source_velocity_mx_mps, temp_range_mx_m)
    else:
        temp_range_mx_m = range_initial_mx_m - source_velocity_mx_mps * spacetime_matrix
        term1 = (signal_speed_mps**2) * time_array_s + \
            hadamard_dot_product_mx(receiver_velocity_mx_mps, temp_range_mx_m)
    term1 *= denom

    temp_range_m = range_hadamard(temp_range_mx_m)
    term2 = temp_range_m**2 - (time_array_s * signal_speed_mps)**2
    term2 *= denom
    if inverse:
        time_s = term1 - np.sqrt(term1**2 + term2)
        range_mx_m = temp_range_mx_m - source_velocity_mx_mps * time_4d_mx(time_s, num_dimensions)
    else:
        time_s = term1 + np.sqrt(term1**2 + term2)
        range_mx_m = temp_range_mx_m + receiver_velocity_mx_mps * time_4d_mx(time_s, num_dimensions)

    range_time_m = range_hadamard(range_mx_m)
    omega_over_omega_center = \
        (signal_speed_mps - hadamard_dot_product_mx(range_mx_m, receiver_velocity_mx_mps) / range_time_m) / \
        (signal_speed_mps - hadamard_dot_product_mx(range_mx_m, source_velocity_mx_mps) / range_time_m)

    return time_s, range_time_m, omega_over_omega_center


def doppler_forward(
        tau_source_s: np.array,
        signal_speed_mps: float,
        source_speed_mps: float,
        receiver_speed_mps: float,
        space_dimensions: int,
        source_position_vector_initial_xyz_m: np.array,
        source_position_vector_final_xyz_m: np.array,
        receiver_position_vector_initial_xyz_m: np.array,
        receiver_position_vector_final_xyz_m: np.array
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Apply a doppler shift on a source moving towards the receiver

    :param tau_source_s: array of source times in seconds
    :param signal_speed_mps: speed of the signal in meters per second
    :param source_speed_mps: speed of the source in meters per second
    :param receiver_speed_mps: speed of the receiver in meters per second
    :param space_dimensions: integer number of dimensions everything is moving in
    :param source_position_vector_initial_xyz_m: position in space where source initially started
    :param source_position_vector_final_xyz_m: position in space where source ends up
    :param receiver_position_vector_initial_xyz_m: position in space where receiver initially started
    :param receiver_position_vector_final_xyz_m: position in space where receiver ends up
    :return: array of receiver times in seconds, Magnitude of range and time, omega over omega center
    """
    tau_number_samples, tau_mx_s, source_trajectory_m, receiver_trajectory_m = \
        _get_setup(tau_source_s, space_dimensions, source_position_vector_initial_xyz_m,
                   source_position_vector_final_xyz_m, receiver_position_vector_initial_xyz_m,
                   receiver_position_vector_final_xyz_m)

    source_velocity_mx_mps = _get_velocity_mps(source_speed_mps, source_trajectory_m, tau_number_samples,
                                               source_position_vector_initial_xyz_m,
                                               source_position_vector_final_xyz_m)

    receiver_velocity_mx_mps = _get_velocity_mps(receiver_speed_mps, receiver_trajectory_m, tau_number_samples,
                                                 receiver_position_vector_initial_xyz_m,
                                                 receiver_position_vector_final_xyz_m)
    # Observation time
    return _get_final_vals(tau_mx_s, receiver_velocity_mx_mps, source_velocity_mx_mps, tau_source_s,
                           receiver_position_vector_initial_xyz_m, source_position_vector_initial_xyz_m,
                           signal_speed_mps, receiver_speed_mps, space_dimensions, tau_number_samples, False)


def image_doppler_forward(
        tau_source_s: np.array,
        signal_speed_mps: float,
        source_speed_mps: float,
        receiver_speed_mps: float,
        space_dimensions: int,
        source_position_vector_initial_xyz_m: np.array,
        source_position_vector_final_xyz_m: np.array,
        receiver_position_vector_initial_xyz_m: np.array,
        receiver_position_vector_final_xyz_m: np.array
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Apply a doppler shift on a perceived source moving towards the receiver

    :param tau_source_s: array of source times in seconds
    :param signal_speed_mps: speed of the signal in meters per second
    :param source_speed_mps: speed of the source in meters per second
    :param receiver_speed_mps: speed of the receiver in meters per second
    :param space_dimensions: integer number of dimensions everything is moving in
    :param source_position_vector_initial_xyz_m: position in space where source initially started
    :param source_position_vector_final_xyz_m: position in space where source ends up
    :param receiver_position_vector_initial_xyz_m: position in space where receiver initially started
    :param receiver_position_vector_final_xyz_m: position in space where receiver ends up
    :return: array of receiver times in seconds, Magnitude of range and time, omega over omega center
    """
    # Flip the z-axis - everything else is the same
    image_source_position_vector_initial_xyz_m = source_position_vector_initial_xyz_m*np.array([1., 1., -1.])
    image_source_position_vector_final_xyz_m = source_position_vector_final_xyz_m*np.array([1., 1., -1.])

    return doppler_forward(tau_source_s, signal_speed_mps, source_speed_mps, receiver_speed_mps,
                           space_dimensions, image_source_position_vector_initial_xyz_m,
                           image_source_position_vector_final_xyz_m, receiver_position_vector_initial_xyz_m,
                           receiver_position_vector_final_xyz_m)


def doppler_inverse(
        inv_time_receiver_s: np.array,
        signal_speed_mps: float,
        source_speed_mps: float,
        receiver_speed_mps: float,
        space_dimensions: int,
        source_position_vector_initial_xyz_m: np.array,
        source_position_vector_final_xyz_m: np.array,
        receiver_position_vector_initial_xyz_m: np.array,
        receiver_position_vector_final_xyz_m: np.array
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Apply a doppler shift on an inverse

    :param inv_time_receiver_s: array of receiver times in seconds
    :param signal_speed_mps: speed of the signal in meters per second
    :param source_speed_mps: speed of the source in meters per second
    :param receiver_speed_mps: speed of the receiver in meters per second
    :param space_dimensions: integer number of dimensions everything is moving in
    :param source_position_vector_initial_xyz_m: position in space where source initially started
    :param source_position_vector_final_xyz_m: position in space where source ends up
    :param receiver_position_vector_initial_xyz_m: position in space where receiver initially started
    :param receiver_position_vector_final_xyz_m: position in space where receiver ends up
    :return: array of receiver times in seconds, Magnitude of range and time, omega over omega center
    """
    time_number_samples, time_mx_s, source_trajectory_m, receiver_trajectory_m = \
        _get_setup(inv_time_receiver_s, space_dimensions, source_position_vector_initial_xyz_m,
                   source_position_vector_final_xyz_m, receiver_position_vector_initial_xyz_m,
                   receiver_position_vector_final_xyz_m)

    source_velocity_mx_mps = _get_velocity_mps(source_speed_mps, source_trajectory_m, time_number_samples,
                                               source_position_vector_initial_xyz_m,
                                               source_position_vector_final_xyz_m)

    receiver_velocity_mx_mps = _get_velocity_mps(receiver_speed_mps, receiver_trajectory_m, time_number_samples,
                                                 receiver_position_vector_initial_xyz_m,
                                                 receiver_position_vector_final_xyz_m)

    # compute range and terms and find Source tau
    return _get_final_vals(time_mx_s, receiver_velocity_mx_mps, source_velocity_mx_mps, inv_time_receiver_s,
                           receiver_position_vector_initial_xyz_m, source_position_vector_initial_xyz_m,
                           signal_speed_mps, source_speed_mps, space_dimensions, time_number_samples, True)


def image_doppler_inverse(
        inv_time_receiver_s: np.array,
        signal_speed_mps: float,
        source_speed_mps: float,
        receiver_speed_mps: float,
        space_dimensions: int,
        source_position_vector_initial_xyz_m: np.array,
        source_position_vector_final_xyz_m: np.array,
        receiver_position_vector_initial_xyz_m: np.array,
        receiver_position_vector_final_xyz_m: np.array
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Apply a doppler shift on an image inverse

    :param inv_time_receiver_s: array of receiver times in seconds
    :param signal_speed_mps: speed of the signal in meters per second
    :param source_speed_mps: speed of the source in meters per second
    :param receiver_speed_mps: speed of the receiver in meters per second
    :param space_dimensions: integer number of dimensions everything is moving in
    :param source_position_vector_initial_xyz_m: position in space where source initially started
    :param source_position_vector_final_xyz_m: position in space where source ends up
    :param receiver_position_vector_initial_xyz_m: position in space where receiver initially started
    :param receiver_position_vector_final_xyz_m: position in space where receiver ends up
    :return: array of receiver times in seconds, Magnitude of range and time, omega over omega center
    """
    # Flip the z-axis - everything else is the same
    image_source_position_vector_initial_xyz_m = source_position_vector_initial_xyz_m*np.array([1., 1., -1.])
    image_source_position_vector_final_xyz_m = source_position_vector_final_xyz_m*np.array([1., 1., -1.])

    return doppler_inverse(inv_time_receiver_s, signal_speed_mps, source_speed_mps, receiver_speed_mps,
                           space_dimensions, image_source_position_vector_initial_xyz_m,
                           image_source_position_vector_final_xyz_m, receiver_position_vector_initial_xyz_m,
                           receiver_position_vector_final_xyz_m)



""""
Standardized Stockwell transform (stx) with optimization parameters
After Moukadem et al., 2022, A new optimized Stockwell transform applied on synthetic and real non-stationary signals
Rederivation in preparation, Garces et al. 2022; last updated 9 May 2022
"""
import numpy as np
from scipy.fft import fft, ifft, fftfreq
from quantum_inferno import scales_dyadic as scales
from typing import Tuple
from quantum_inferno.utilities.rescaling import is_power_of_two


def sig_pad_up_to_pow2(sig_wf: np.ndarray, n_fft: int, verbosity: bool = False):
    """
    Zero-pad signal to the higher 2**n points for FFT

    :param sig_wf: waveform to add zero padding
    :param n_fft: length of fft
    :param verbosity: if True, outputs additional debugging statements
    :return: waveform with zero padding, length of fft to nearest greater power of 2, number of points added
    """
    # Flatten to 2D and memorize original shape
    n_times = sig_wf.shape[-1]

    if verbosity:
        print(f"length of fft: {n_fft}, waveform shape: {n_times}")
    if n_fft < n_times:
        raise ValueError(f"n_fft cannot be smaller than signal size. Got {n_fft} < {n_times}.")

    if n_fft is None or (not is_power_of_two(n_fft) and n_times > n_fft):
        # Compute next power of 2
        n_fft = 2 ** int(np.ceil(np.log2(n_times)))

    if n_times < n_fft:
        if verbosity:
            print(
                f'The input signal is shorter ({sig_wf.shape[-1]}) than "n_fft" ({n_fft}). '
                "Applying zero padding."
            )
        zero_pad: int = n_fft - n_times
        sig_wf = np.concatenate((sig_wf, np.zeros(sig_wf.shape[:-1] + zero_pad, sig_wf.dtype)), axis=-1)
    else:
        zero_pad: int = 0

    return sig_wf, n_fft, zero_pad


# Taken from the latest libquantum version, adjusted for use with quantum_inferno
def tfr_stx_fft(
    sig_wf: np.ndarray,
    time_sample_interval: float,
    scale_order_input: float = 8.0,
    n_fft_in: int = None,
    frequency_min: float = None,
    frequency_max: float = None,
    frequency_step: float = None,
    factor_q: float = 0.0,
    power_p: float = 0.0,
    power_r: float = 1.0,
    is_geometric: bool = False,
    is_inferno: bool = False,
    scale_base_input: float = scales.Slice.G3,
    scale_ref_input: float = scales.Slice.T1S,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Stockwell transform, fft implementation.
    Optimized version has more free variables in sigma_scaling; testing in progress

    :param sig_wf: input waveform. If not 2**n points, it will zero pad up
    :param time_sample_interval: sample interval, inverse of sample rate
    :param scale_order_input: fractional octave band order; 12 is the musical standard
    :param n_fft_in: requested nfft, should be greater or equal to the signal length. Method zero pads up.
    :param frequency_min: lowest stx frequency of interest
    :param frequency_max: highest stx frequency of interest
    :param frequency_step: stx frequency of interest if linearly sampled
    :param factor_q: sigma_scaling adjustment, under evaluation
    :param power_p: sigma_scaling adjustment, under evaluation
    :param power_r: sigma_scaling adjustment. under evaluation
    :param is_geometric: are frequencies geometrically spaced? If so, overrides frequency_step
    :param is_inferno: are frequencies geometrically spaced and standardized as in inferno?
    :param scale_base_input: scale base; default is base 10, base 2 octaves is scales.Slice.G2
    :param scale_ref_input: scale reference time; default is 1 s
    :return: tfr_stx, psd_stx, frequency_stx_hz, frequency_stx_fft, windows_fft
    """
    frequency_sample_rate: float = 1 / time_sample_interval
    cycles_m: float = 12.0 / 5.0 * scale_order_input
    lin_fft_decimate: float = 2.0

    # Compute the nearest higher power of two number of points for the fft
    sig_wf_pow2, n_fft_pow2, zero_pad = sig_pad_up_to_pow2(sig_wf, n_fft_in)

    # For reduction back to the near original signal after taking the ifft. Should match input sig.
    n_fft_out = n_fft_pow2 - zero_pad

    # Transformations are on zero padded signals from here onwards
    # Take FFT and concatenate. A leaner version could let the fft do the padding.
    sig_fft = fft(sig_wf_pow2)
    sig_fft_cat = np.concatenate([sig_fft, sig_fft], axis=-1)

    frequency_fft = fftfreq(n_fft_pow2, time_sample_interval)  # in units of 1/sample interval
    omega_fft = 2 * np.pi * frequency_fft / frequency_sample_rate  # scaled angular frequency

    window_longest_time = n_fft_pow2 / frequency_sample_rate
    frequency_min_nth = cycles_m / window_longest_time

    # Initialize stx frequencies
    if frequency_min is None:
        frequency_min = frequency_min_nth
    if frequency_max is None:
        frequency_max = frequency_sample_rate / 2.0

    # Computing nearest frequency later on anyway, and then using that to compute the fft.
    start_f_idx = np.abs(frequency_fft - frequency_min).argmin()
    stop_f_idx = np.abs(frequency_fft - frequency_max).argmin()
    f_start = frequency_fft[start_f_idx]
    f_stop = frequency_fft[stop_f_idx]

    # Linear scale
    if frequency_step is None:
        # Reduce the fft resolution by a factor of lin_fft_decimate
        frequency_step = (frequency_max - frequency_min) * lin_fft_decimate / len(frequency_fft)
        frequency_stx = np.arange(f_start, f_stop, frequency_step)
    else:
        frequency_stx = np.arange(f_start, f_stop, frequency_step)

    # if geometric (log) frequency
    if is_geometric is True:
        # if standardized to ISO3
        if is_inferno is True:
            (
                order_Nth,
                scale_base,
                scale_band_number,
                frequency_ref,
                frequency_center_algebraic,
                frequency_center_geometric,
                frequency_start,
                frequency_end,
            ) = scales.band_frequency_low_high(
                frequency_order_input=scale_order_input,
                frequency_low_input=f_start,
                frequency_high_input=f_stop,
                frequency_sample_rate_input=frequency_sample_rate,
                frequency_base_input=scale_base_input,
                frequency_ref_input=scale_ref_input,
            )
            frequency_stx = frequency_center_geometric
        else:
            num_octaves = np.log2(f_stop / f_start)
            num_bands = int(num_octaves * scale_order_input)
            frequency_stx = np.logspace(np.log2(f_start), np.log2(f_stop), num=num_bands, base=scale_base_input)

    # Construct shifting frequency indexes
    frequency_stx_fft = np.empty(len(frequency_stx))

    # Construct time domain and fft of window
    windows_fft = np.empty((len(frequency_stx), n_fft_pow2), dtype=np.complex128)
    # tfr_stx_pow2 = np.empty(1, n_fft_pow2, dtype=np.complex128)

    tfr_stx = np.empty((len(frequency_stx), n_fft_out), dtype=np.complex128)
    psd_stx = np.empty((len(frequency_stx), n_fft_out))

    # Minimized for loops and operations
    for isx, fsx in enumerate(frequency_stx):
        stx_index = np.abs(frequency_fft - fsx).argmin()
        frequency_stx_fft[isx] = frequency_fft[stx_index]
        # eq non-dimensional angular fft frequency
        omega_sx = 2 * np.pi * frequency_stx_fft[isx] / frequency_sample_rate
        if omega_sx == 0.0:
            windows_fft[isx] = np.ones(len(n_fft_pow2))
        else:
            # Sigma is the standard deviation of the Gaussian
            sigma_scaling = (1 + factor_q * (omega_sx ** power_p)) * (omega_sx ** (1 - power_r))
            sigma = cycles_m / omega_sx
            sigma *= sigma_scaling
            windows_fft[isx] = np.exp(-0.5 * (sigma ** 2.0) * (omega_fft ** 2.0))

        # This is the main event
        tfr_stx_pow2 = ifft(sig_fft_cat[stx_index: stx_index + n_fft_pow2] * windows_fft[isx])
        if zero_pad > 0:
            tfr_stx[isx, :] = tfr_stx_pow2[:-zero_pad:1]
        else:
            tfr_stx[isx, :] = tfr_stx_pow2
        # Power
        tfr_abs = np.abs(tfr_stx[isx, :]) ** 2
        # Add EPSILON for zero handling
        psd_stx[isx, :] = tfr_abs + scales.get_epsilon()

    return tfr_stx, psd_stx, frequency_stx, frequency_stx_fft, windows_fft


def stx_complex_any_scale_pow2(
    band_order_nth: float, sig_wf: np.ndarray, frequency_sample_rate_hz: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    With some assumptions and simplifications, and with some vectorization

    :param band_order_nth: Fractional octave band - revisit
    :param sig_wf: input signal with 2^M points
    :param frequency_sample_rate_hz: sample rate in Hz
    :return: frequency_stx_hz, time_stx_s, tfr_stx
    """
    n_fft_pow2 = len(sig_wf)
    frequency_stx_hz = scales.log_frequency_hz_from_fft_points(
        frequency_sample_hz=frequency_sample_rate_hz, fft_points=n_fft_pow2, scale_order=band_order_nth
    )

    scale_points = len(frequency_stx_hz)
    # Take FFT and concatenate. A leaner version could let the fft do the padding.
    sig_fft = fft(sig_wf)
    sig_fft_cat = np.concatenate([sig_fft, sig_fft], axis=-1)

    frequency_fft = fftfreq(n_fft_pow2, 1 / frequency_sample_rate_hz)  # in units of 1/sample interval
    omega_fft = 2 * np.pi * frequency_fft / frequency_sample_rate_hz  # scaled angular frequency
    omega_stx = 2 * np.pi * frequency_stx_hz / frequency_sample_rate_hz  # non-dimensional angular stx frequency
    sigma_stx = scales.cycles_from_order(scale_order=band_order_nth) / omega_stx

    # Construct 2d matrices
    # sig_fft_cat_2d = np.tile(sig_fft_cat, (scale_points, 1))
    windows_fft_2d = np.exp(
        -0.5 * (np.tile(sigma_stx, (n_fft_pow2, 1)).T ** 2.0) * (np.tile(omega_fft, (scale_points, 1)) ** 2.0)
    )

    # Construct shifting frequency indexes
    tfr_stx = np.empty((scale_points, n_fft_pow2), dtype=np.complex128)

    # Minimal iteration
    for isx, fsx in enumerate(frequency_stx_hz):
        # This is the main event
        stx_index = np.abs(frequency_fft - fsx).argmin()
        tfr_stx[isx, :] = ifft(sig_fft_cat[stx_index: stx_index + n_fft_pow2] * windows_fft_2d[isx, :])

    return frequency_stx_hz, np.arange(n_fft_pow2) / frequency_sample_rate_hz, tfr_stx



"""
Methods for creating windows.
"""

import numpy as np
from scipy.signal import windows


def get_tukey(array: np.ndarray, alpha: float = 0.5) -> np.ndarray:
    """
    Create a symmetric Tukey window (AKA: tapered cosine window) with same shape as input array.
    Note: alpha of 0 is a rectangular window, 1 is a Hann window.
    :param array: input array to get shape from
    :param alpha: fraction of the window inside the cosine tapered window, shared between the head and tail
    """
    return windows.tukey(M=np.size(array), alpha=alpha, sym=True)


def get_tukey_by_buffer_num(array: np.ndarray, taper_num: int, alpha: float = 0.5) -> np.ndarray:
    """
    Create window with the same shape as input array, tukey tapered on the head and tail with ones in the middle.

    Note: alpha of 0 is a rectangular window, 1 is a Hann window.
    :param array: input array to get shape from
    :param taper_num: number of points to taper on each side
    :param alpha: fraction of the window inside the cosine tapered window, shared between the head and tail
    """
    if len(array) < taper_num * 2:
        print(f"Warning: array length {len(array)} is less than taper_num {taper_num * 2}. Using full array length.")
        return get_tukey(array, alpha=alpha)

    tukey = windows.tukey(taper_num * 2, alpha=alpha, sym=True)
    middle_ones = np.ones(len(array) - taper_num * 2)
    return np.concatenate((tukey[:taper_num], middle_ones, tukey[taper_num:]))


def get_tukey_by_buffer_s(array: np.ndarray, taper_s: float, sample_rate_hz: float, alpha: float = 0.5) -> np.ndarray:
    """
    Create window with the same shape as input array, tukey tapered on the head and tail with ones in the middle.

    Note: alpha of 0 is a rectangular window, 1 is a Hann window.
    :param array: input array to get shape from
    :param taper_s: duration of taper in seconds
    :param sample_rate_hz: sample rate in Hz
    :param alpha: fraction of the window inside the cosine tapered window, shared between the head and tail
    """
    taper_num = int(taper_s * sample_rate_hz)
    return get_tukey_by_buffer_num(array, taper_num, alpha=alpha)



import unittest

import numpy as np

import quantum_inferno.plot_templates.plot_templates as pt


class SanitizeTimestampsTest(unittest.TestCase):
    def test_sanitize_timestamps(self):
        timestamps = np.array([10, 20, 30, 40, 50])
        new_timestamps = pt.sanitize_timestamps(timestamps)
        self.assertEqual(new_timestamps[0], 0)


class GetTimeLabelTest(unittest.TestCase):
    def test_get_time_label(self):
        start_time_epoch = 1704067200  # 2024 Jan 1, 00:00
        time_units = "s"
        time_label = pt.get_time_label(start_time_epoch, time_units, 0.)
        self.assertEqual(time_label, "Time (s) from UTC 2024-01-01 00:00:00")


class MeshTimeFrequencyEdgesTest(unittest.TestCase):
    def test_mesh_time_freq_edge(self):
        frequency = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        timestamps = np.array([0, 10, 20, 30, 40, 50, 60, 70, 80, 90])
        min_freq = 0
        max_freq = 9

        t_edge, f_edge, f_min, f_max = pt.mesh_time_frequency_edges(frequency, timestamps, min_freq, max_freq)

        self.assertEqual(len(t_edge), len(timestamps) + 1)
        self.assertEqual(t_edge[0], -5.)
        self.assertEqual(len(f_edge), len(frequency) + 1)
        self.assertEqual(f_edge[0], -.5)
        self.assertEqual(f_min, -.5)
        self.assertEqual(f_max, 9.)



"""
This module contains methods used in resampling signals
"""
from typing import Tuple

import numpy as np
from scipy.signal import resample, decimate


SUBSAMPLE_METHODS = ["average", "median", "max", "min", "nth"]


def subsample(
    timeseries: np.ndarray, sample_rate_hz: float, subsample_factor: int, method: str = "nth"
) -> Tuple[np.ndarray, float]:
    """
    Subsample a time series by given method (default is every nth sample).
    Truncates the signal if the length is not divisible by the subsample factor, except for the nth method.
    if subsample_factor is less than 2, return the original signal.

    :param timeseries: input signal
    :param sample_rate_hz: sample rate of the signal
    :param subsample_factor: factor to subsample by (returns every nth sample)
    :param method: method to use for subsampling
    :return: subsampled signal and new sample rate
    """
    if subsample_factor < 2:
        print(f"Warning: subsample factor is less than 2, returning the original signal")
        return timeseries, sample_rate_hz

    new_sample_rate = sample_rate_hz / subsample_factor

    if method not in SUBSAMPLE_METHODS:
        print(f"Warning: method {method} not recognized, using 'nth' method")
        method = "nth"

    if method != "nth" and len(timeseries) % subsample_factor != 0:
        timeseries = timeseries[: -(len(timeseries) % subsample_factor)]

    if method == "nth":
        return timeseries[::subsample_factor], new_sample_rate
    elif method == "average":
        return np.mean(timeseries.reshape(-1, subsample_factor), axis=1), new_sample_rate
    elif method == "median":
        return np.median(timeseries.reshape(-1, subsample_factor), axis=1), new_sample_rate
    elif method == "max":
        return np.max(timeseries.reshape(-1, subsample_factor), axis=1), new_sample_rate
    elif method == "min":
        return np.min(timeseries.reshape(-1, subsample_factor), axis=1), new_sample_rate


def resample_uneven_timeseries(
    timeseries: np.ndarray, timestamps_s: np.ndarray, new_sample_rate_hz: float or None = None
) -> Tuple[np.ndarray, float]:
    """
    Resample uneven time series using linear interpolation.
    If new_sample_rate_hz is None, the new sample rate is the average sample rate of the input signal.

    :param timeseries: input signal
    :param timestamps_s: timestamps of the input signal in seconds
    :param new_sample_rate_hz: the sample rate to resample to in Hz (default is None)
    :return: resampled signal and new sample rate
    """
    if new_sample_rate_hz is None:
        new_sample_rate_hz = 1 / np.mean(np.diff(timestamps_s))
    new_timestamps = np.arange(timestamps_s[0], timestamps_s[-1], 1 / new_sample_rate_hz)
    return np.interp(new_timestamps, timestamps_s, timeseries), new_sample_rate_hz


def resample_with_sample_rate(
    timeseries: np.ndarray, sample_rate_hz: float, new_sample_rate_hz: float
) -> Tuple[np.ndarray, float]:
    """
    Resample a time series to a new sample rate using scipy.signal.resample.

    :param timeseries: input signal
    :param sample_rate_hz: sample rate of the input signal
    :param new_sample_rate_hz: the sample rate to resample to in Hz
    :return: resampled signal and new sample rate
    """
    new_length = int(len(timeseries) * new_sample_rate_hz / sample_rate_hz)
    return resample(timeseries, new_length), new_sample_rate_hz


# subsample a 2d array along the second axis
def subsample_2d(array: np.ndarray, subsample_factor: int, method: str = "nth") -> np.ndarray:
    """
    Subsample a 2D array along the second axis.
    Truncates the signal if the length is not divisible by the subsample factor.
    if subsample_factor is less than 2, return the original signal.

    :param array: input 2D array
    :param subsample_factor: factor to subsample
    :param method: method to use for subsampling (default is every nth sample)
    :return: subsampled 2D array
    """
    if subsample_factor < 2:
        print(f"Warning: subsample factor is less than 2, returning the original signal")
        return array

    if method not in SUBSAMPLE_METHODS:
        print(f"Warning: method {method} not recognized, using 'nth' method")
        method = "nth"

    if method != "nth":
        remainder = array.shape[1] % subsample_factor
        if remainder != 0:
            array = array[:, :-remainder]

    if method == "nth":
        return array[:, ::subsample_factor]
    elif method == "average":
        return np.mean(array.reshape(-1, subsample_factor), axis=1).reshape(array.shape[0], -1)
    elif method == "median":
        return np.median(array.reshape((-1, subsample_factor)), axis=1).reshape(array.shape[0], -1)
    elif method == "max":
        return np.max(array.reshape((-1, subsample_factor)), axis=1).reshape(array.shape[0], -1)
    elif method == "min":
        return np.min(array.reshape((-1, subsample_factor)), axis=1).reshape(array.shape[0], -1)


# decimate a 1d array
def decimate_timeseries(timeseries: np.ndarray, decimation_factor: int) -> np.ndarray:
    """
    Decimate a time series by a given factor using scipy.signal.decimate.
    Timeseries must be 28 samples or longer.

    :param timeseries: input signal
    :param decimation_factor: factor to decimate by
    :return: decimated signal
    """
    return decimate(timeseries, decimation_factor, zero_phase=True)


# decimate a collection of timeseries with the same sample rate at once
def decimate_timeseries_collection(timeseries_collection: np.ndarray, decimation_factor: int) -> np.ndarray:
    """
    Decimate a collection of time series with the same sample rate at once using scipy.signal.decimate.
    Each timeseries must be 28 samples or longer.

    :param timeseries_collection: input signal
    :param decimation_factor: factor to decimate by
    :return: decimated signal
    """
    return decimate(timeseries_collection, decimation_factor, axis=1, zero_phase=True)



import unittest

import numpy as np
from scipy.signal import ShortTimeFFT

from quantum_inferno.synth.benchmark_signals import well_tempered_tone
from quantum_inferno.utilities import short_time_fft


class TestPicker(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.tukey_alpha = 0.25
        cls.signal, cls.timestamps, cls.fft_nd, cls.sample_rate, cls.freq_center, cls.resolution = well_tempered_tone()
        cls.two_peak_signal = np.array([0, 1, 2, 3, 4, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -4, -3, -2, -1])

    def test_get_stft_object_tukey(self):
        stft_obj = short_time_fft.get_stft_object_tukey(
            sample_rate_hz=self.sample_rate,
            tukey_alpha=self.tukey_alpha,
            segment_length=self.fft_nd,
            overlap_length=self.fft_nd // 2,
            scaling="magnitude",
        )
        self.assertIsInstance(stft_obj, ShortTimeFFT)
        self.assertTrue(stft_obj.invertible)
        self.assertEqual(stft_obj.scaling, "magnitude")
        self.assertEqual(stft_obj.fft_mode, "onesided")
        self.assertEqual(stft_obj.m_num, self.fft_nd)  # check window length
        self.assertEqual(stft_obj.mfft, self.fft_nd)  # check window length with mfft
        self.assertEqual(stft_obj.hop, self.fft_nd // 2)  # check hop length
        self.assertEqual(stft_obj.delta_f, self.resolution)

    def test_stft_tukey(self):
        frequencies, times, stfts = short_time_fft.stft_tukey(
            timeseries=self.signal,
            sample_rate_hz=self.sample_rate,
            tukey_alpha=self.tukey_alpha,
            segment_length=self.fft_nd,
            overlap_length=self.fft_nd // 2,
            scaling="magnitude",
        )
        self.assertEqual(times.shape[0], len(self.timestamps) // (self.fft_nd // 2) + 1)
        self.assertEqual(frequencies.shape, (self.fft_nd // 2 + 1,))
        self.assertEqual(stfts.shape, (self.fft_nd // 2 + 1, len(self.timestamps) // (self.fft_nd // 2) + 1))

    def test_istft_tukey(self):
        frequencies, times, stfts = short_time_fft.stft_tukey(
            timeseries=self.signal,
            sample_rate_hz=self.sample_rate,
            tukey_alpha=self.tukey_alpha,
            segment_length=self.fft_nd,
            overlap_length=self.fft_nd // 2,
            scaling="magnitude",
        )
        reconstructed_time, reconstructed_signal = short_time_fft.istft_tukey(
            stft_to_invert=stfts,
            sample_rate_hz=self.sample_rate,
            tukey_alpha=self.tukey_alpha,
            segment_length=self.fft_nd,
            overlap_length=self.fft_nd // 2,
            scaling="magnitude",
        )
        self.assertEqual(len(reconstructed_signal), len(self.signal))
        self.assertTrue(np.allclose(self.signal, reconstructed_signal, atol=1e-14))
        self.assertTrue(np.allclose(self.timestamps, reconstructed_time, atol=1e-14))

    def test_spectrogram_tukey(self):
        frequencies, times, spectrogram = short_time_fft.spectrogram_tukey(
            timeseries=self.signal,
            sample_rate_hz=self.sample_rate,
            tukey_alpha=self.tukey_alpha,
            segment_length=self.fft_nd,
            overlap_length=self.fft_nd // 2,
            scaling="magnitude",
        )
        self.assertEqual(times.shape[0], len(self.timestamps) // (self.fft_nd // 2) + 1)
        self.assertEqual(frequencies.shape, (self.fft_nd // 2 + 1,))
        self.assertEqual(spectrogram.shape, (self.fft_nd // 2 + 1, len(self.timestamps) // (self.fft_nd // 2) + 1))



# import unittest
# import numpy as np
# import quantum_inferno.scales_dyadic as sd
#
#
# class MyTestCase(unittest.TestCase):
#     def test_scales_dyadic(self):
#         scale_order0 = 6.
#         frequency_sample_hz0 = 100.
#         time_display_points_float: float = sd.DEFAULT_TIME_DISPLAY_S*frequency_sample_hz0
#         time_display_points_pow2: int = int(2**np.ceil(np.log2(time_display_points_float)))
#         physical_frequency_hz = \
#             sd.log_frequency_hz_from_fft_points(frequency_sample_hz0, time_display_points_pow2,
#                                                 scale_order0, sd.Slice.F1HZ, sd.Slice.G3)
#
#         self.assertEqual(6000., time_display_points_float)
#         self.assertEqual(8192, time_display_points_pow2)
#         self.assertEqual(13.0, np.log2(time_display_points_pow2))
#         self.assertEqual(0.1778279410038923, physical_frequency_hz[0])
#         self.assertEqual(39.810717055349706, physical_frequency_hz[-1])
#         self.assertEqual(48, len(physical_frequency_hz))



"""
Tests for the plot_base module.
"""
import unittest

import numpy as np

import quantum_inferno.plot_templates.plot_base as pb


class PlotBaseTest(unittest.TestCase):
    def test_plot_base(self):
        test_plot_base = pb.PlotBase(station_id="test", figure_title="test")
        self.assertEqual(test_plot_base.station_id, "test")
        self.assertEqual(test_plot_base.figure_title, "test")
        self.assertTrue(test_plot_base.figure_title_show)
        self.assertEqual(test_plot_base.start_time_epoch, 0.)
        self.assertEqual(test_plot_base.units_time, "s")


class MeshBaseTest(unittest.TestCase):
    def test_mesh_base(self):
        test_mesh_base = pb.MeshBase(time=np.array([1, 2, 3]), frequency=np.array([1, 2, 3]))
        self.assertEqual(test_mesh_base.frequency_scaling, "log")
        self.assertEqual(test_mesh_base.shading, "auto")
        self.assertEqual(test_mesh_base.frequency_hz_ymin, 1)
        self.assertEqual(test_mesh_base.frequency_hz_ymax, 3)
        self.assertIsNone(test_mesh_base.colormap)
        self.assertEqual(test_mesh_base.units_frequency, "Hz")

    def test_mesh_base_freq_scal(self):
        test_mesh_base = pb.MeshBase(time=np.array([1, 2, 3]), frequency=np.array([1, 2, 3]),
                                     frequency_scaling="test")
        self.assertEqual(test_mesh_base.frequency_scaling, "log")

    def test_mesh_base_shading(self):
        test_mesh_base = pb.MeshBase(time=np.array([1, 2, 3]), frequency=np.array([1, 2, 3]),
                                     shading="test")
        self.assertEqual(test_mesh_base.shading, "auto")


class MeshColormapLimitsTest(unittest.TestCase):
    def test_mesh_colormap_limits(self):
        mesh_array = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 0])
        test_min, test_max = pb.mesh_colormap_limits(mesh_array)
        self.assertEqual(test_min, 0.)
        self.assertEqual(test_max, 9.)

    def test_mesh_colormap_limits_short(self):
        mesh_array = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 0])
        test_min, test_max = pb.mesh_colormap_limits(mesh_array, colormap_scaling="range", color_range=4)
        self.assertEqual(test_min, 5.)
        self.assertEqual(test_max, 9.)

    def test_mesh_colormap_limits_long(self):
        mesh_array = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 0])
        test_min, test_max = pb.mesh_colormap_limits(mesh_array, colormap_scaling="range", color_range=12)
        self.assertEqual(test_min, -3.)
        self.assertEqual(test_max, 9.)

    def test_mesh_colormap_limits_abs_scaling(self):
        mesh_array = np.array([-1, 2, 3, 4, -5, 6, 7, 8, -9, 0])
        test_min, test_max = pb.mesh_colormap_limits(mesh_array, colormap_scaling="abs")
        self.assertEqual(test_min, 0.)
        self.assertEqual(test_max, 9.)


class MeshPanelTest(unittest.TestCase):
    def test_mesh_panel(self):
        test_mesh_panel = pb.MeshPanel(tfr=np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 0]))
        self.assertTrue(test_mesh_panel.is_auto_color_min_max())

    def test_mesh_user_panel(self):
        test_mesh_panel = pb.MeshPanel(tfr=np.array([1, 2, 3]), colormap_scaling="user", color_min=5, color_max=10)
        self.assertFalse(test_mesh_panel.is_auto_color_min_max())
        self.assertEqual(test_mesh_panel.color_min, 5)
        self.assertEqual(test_mesh_panel.color_max, 10)

    def test_mesh_ytick_style(self):
        test_mesh_panel = pb.MeshPanel(tfr=np.array([1, 2, 3]), ytick_style="test")
        self.assertEqual(test_mesh_panel.ytick_style, "sci")


class WaveformBaseTest(unittest.TestCase):
    def test_waveform_base(self):
        test_wf_base = pb.WaveformPlotBase(station_id="test", figure_title="test")
        self.assertFalse(test_wf_base.label_panel_show)
        self.assertIsNone(test_wf_base.labels_fontweight)
        self.assertIsNone(test_wf_base.waveform_color)


class WaveformPanelTest(unittest.TestCase):
    def test_waveform_panel(self):
        test_wf_panel = pb.WaveformPanel(sig=np.array([1, 2, 3]), time=np.array([1, 2, 3]))
        self.assertEqual(test_wf_panel.units, "Norm")
        self.assertEqual(test_wf_panel.label, "(wf)")
        self.assertEqual(test_wf_panel.yscaling, "auto")
        self.assertEqual(test_wf_panel.ytick_style, "plain")

    def test_waveform_panel_yscaling(self):
        test_wf_panel = pb.WaveformPanel(sig=np.array([1, 2, 3]), time=np.array([1, 2, 3]), yscaling="test")
        self.assertEqual(test_wf_panel.yscaling, "else")

    def test_waveform_panel_ytick_style(self):
        test_wf_panel = pb.WaveformPanel(sig=np.array([1, 2, 3]), time=np.array([1, 2, 3]), ytick_style="test")
        self.assertEqual(test_wf_panel.ytick_style, "plain")



"""
Collection of functions to convert between time bases
"""

from datetime import datetime, timezone
from typing import Union

import numpy as np

# dictionary of time units and their conversion factors to seconds (can add more units as needed)
time_unit_dict = {
    "ps": 1e-12,  # "picosecond"
    "ns": 1e-9,
    "us": 1e-6,
    "ms": 1e-3,
    "s": 1,
    "m": 60,
    "h": 3600,
    "d": 86400,
    "weeks": 604800,
    "months": 2628000,
    "years": 31536000,
}


def convert_time_unit(
    input_time: Union[np.ndarray, float], input_unit: str, output_unit: str
) -> Union[np.ndarray, float]:
    """
    Convert time data from a given time unit to another time unit.

    :param input_time: time data to convert
    :param input_unit: time unit of the input data
    :param output_unit: time unit to convert the input data to
    :return: converted time data
    """
    if input_unit not in time_unit_dict.keys() or output_unit not in time_unit_dict.keys():
        raise ValueError(f"Invalid time unit, please use one of the following: {time_unit_dict.keys()}")
    return input_time * time_unit_dict[input_unit] / time_unit_dict[output_unit]


def utc_datetime_to_utc_timestamp(datetime_obj: datetime, output_unit: str = "s") -> float:
    """
    Convert a UTC datetime object to a UTC timestamp.
    If datetime_object is not timezone aware, it will be assumed to be in UTC.
    If datetime_object is timezone aware, it will be converted to UTC.

    :param datetime_obj: UTC datetime object to convert
    :param output_unit: time unit to convert the UTC timestamp to (default: seconds)
    :return: converted UTC timestamp
    """
    if output_unit not in time_unit_dict.keys():
        raise ValueError(f"Invalid time unit, please use one of the following: {time_unit_dict.keys()}")
    if datetime_obj.tzinfo is None:
        datetime_obj = datetime_obj.replace(tzinfo=timezone.utc)
    elif datetime_obj.tzinfo != timezone.utc:
        datetime_obj = datetime_obj.astimezone(timezone.utc)
    return convert_time_unit(datetime_obj.timestamp(), "s", output_unit)


def utc_timestamp_to_utc_datetime(timestamp: float, input_unit: str = "s") -> datetime:
    """
    Convert a UTC timestamp to a UTC datetime object.
    Note: timestamp is assumed to be in UTC.

    :param timestamp: UTC timestamp to convert
    :param input_unit: time unit of the UTC timestamp (default: seconds)
    :return: converted UTC datetime object
    """
    if input_unit not in time_unit_dict.keys():
        raise ValueError(f"Invalid time unit, please use one of the following: {time_unit_dict.keys()}")
    return datetime.utcfromtimestamp(convert_time_unit(timestamp, input_unit, "s")).replace(tzinfo=timezone.utc)


def set_datetime_to_utc(datetime_obj: datetime, tzinfo_warning: bool = False) -> datetime:
    """
    Convert a datetime object to a UTC datetime object.
    If the input datetime object is not timezone-aware, it is assumed to be in UTC.

    :param datetime_obj: datetime object to convert
    :param tzinfo_warning: flag to raise a warning if the input datetime object is not timezone-aware
    :return: converted UTC datetime object
    """
    if datetime_obj.tzinfo is None:
        if tzinfo_warning:
            print("Warning: input datetime object is not timezone-aware, assuming UTC...")
        return datetime_obj.replace(tzinfo=timezone.utc)
    return datetime_obj.astimezone(timezone.utc)


def set_timestamp_to_utc(timestamp: float, utc_offset_h: float, input_unit: str = "s") -> float:
    """
    Convert a timestamp to be in UTC using the UTC offset.

    :param timestamp: timestamp to convert
    :param utc_offset_h: UTC offset of the timestamp in hours
    :param input_unit: time unit of the timestamp (default: seconds)
    :return: converted timestamp in UTC while keeping the same unit
    """
    if input_unit not in time_unit_dict.keys():
        raise ValueError(f"Invalid time unit, please use one of the following: {time_unit_dict.keys()}")
    offset_in_input_unit = utc_offset_h * time_unit_dict["h"] / time_unit_dict[input_unit]
    return timestamp - offset_in_input_unit


def get_datetime_from_timestamp_to_utc(timestamp: float, utc_offset_h: float, input_unit: str = "s") -> datetime:
    """
    Convert a timestamp to a UTC datetime object using the UTC offset.

    :param timestamp: timestamp to convert into UTC time
    :param utc_offset_h: UTC offset of the timestamp in hours
    :param input_unit: time units (default: seconds)
    :return: converted UTC datetime object
    """
    if input_unit not in time_unit_dict.keys():
        raise ValueError(f"Invalid time unit, please use one of the following: {time_unit_dict.keys()}")
    return utc_timestamp_to_utc_datetime(set_timestamp_to_utc(timestamp, utc_offset_h, input_unit))



"""
This module facilitates the rapid construction of the GT blast pulse synthetic,
its integral and derivatives, and its spectrum
References:
- Garcés, M. A. (2019). Explosion Source Models,
Chapter in Infrasound Monitoring for Atmospheric Studies,
Second Edition, Springer, Switzerland, DOI 10.1007/978-3-319-75140_5, p. 273-345.
- Schnurr, J. M., K. Kim, M. A. Garcés, A. Rodgers (2020).
Improved parametric models for explosion pressure signals derived from large datasets,
Seism. Res. Lett.
- Kim, K, A. R. Rodgers, M. A. Garces, and S. C. Myers (2021).
Empirical Acoustic Source Model for Chemical Explosions in Air.
Bulletin of the Seismological Society of America
"""
from typing import Optional, Tuple, Union

import numpy as np

from quantum_inferno.synth.synthetic_signals import white_noise_fbits, antialias_half_nyquist
from quantum_inferno.scales_dyadic import get_epsilon


def gt_blast_period_center(time_center_s: np.ndarray, pseudo_period_s: float) -> np.ndarray:
    """
    GT blast pulse

    :param time_center_s: array with time
    :param pseudo_period_s: period in seconds
    :return: numpy array with GT blast pulse
    """
    # With the +1, tau is the first zero crossing time.
    time_pos_s = pseudo_period_s / 4.0
    tau = time_center_s / time_pos_s + 1.0
    # Initialize GT
    p_gt = np.zeros(tau.size)  # Granstrom-Triangular (GT), 2019
    # Initialize time ranges
    sigint1 = np.where((0.0 <= tau) & (tau <= 1.0))  # ONLY positive pulse
    sigint_g17 = np.where((1.0 < tau) & (tau <= 1 + np.sqrt(6.0)))  # GT balanced pulse
    p_gt[sigint1] = 1.0 - tau[sigint1]
    p_gt[sigint_g17] = 1.0 / 6.0 * (1.0 - tau[sigint_g17]) * (1.0 + np.sqrt(6) - tau[sigint_g17]) ** 2.0

    return p_gt


def gt_hilbert_blast_period_center(time_center_s: np.ndarray, pseudo_period_s: float) -> np.ndarray:
    """
    Hilbert transform of the GT blast pulse

    :param time_center_s: array with time
    :param pseudo_period_s: period in seconds
    :return: numpy array with Hilbert transform of the GT blast pulse
    """
    # With the +1, tau is the first zero crossing time.
    time_pos_s = pseudo_period_s / 4.0
    tau = time_center_s / time_pos_s + 1.0
    a = 1 + np.sqrt(6)
    # Initialize GT
    p_gt_h = np.zeros(tau.size)  # Hilbert of Granstrom-Triangular (GT), 2019
    # Initialize time ranges
    sigint1 = np.where((0.0 <= tau) & (tau <= 1.0))  # ONLY positive pulse
    sigint2 = np.where((1.0 < tau) & (tau <= 1 + np.sqrt(6.0)))  # GT balanced pulse
    tau1 = tau[sigint1]
    tau2 = tau[sigint2]

    p_gt_h[sigint1] = 1.0 + (1 - tau1) * np.log(tau1 + get_epsilon()) - (1 - tau1) * np.log(1 - tau1 + get_epsilon())
    p_gt_h21 = (a - 1) / 6.0 * (a * (2 * a + 5) - 1 + 6 * tau2 ** 2 - 3 * tau2 * (1 + 3 * a))
    p_gt_h22 = (tau2 - 1) * (a - tau2) ** 2 * (np.log(a - tau2 + get_epsilon()) - np.log(tau2 - 1 + get_epsilon()))
    p_gt_h[sigint2] = 1.0 / 6.0 * (p_gt_h21 + p_gt_h22)
    p_gt_h /= np.pi

    return p_gt_h


def gt_blast_center_fast(
    frequency_peak_hz: float = 6.3, sample_rate_hz: float = 100.0, noise_std_loss_bits: float = 16.
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Fast computation of GT pulse with noise

    :param frequency_peak_hz: peak frequency, nominal 6.3 Hz for 1 tonne TNT
    :param sample_rate_hz: sample rate, nominal 100 Hz
    :param noise_std_loss_bits: noise loss relative to signal variance
    :return: centered time in seconds, GT pulse with white noise
    """
    # 16 cycles for 6th octave (M = 14)
    duration_points = int(16 / frequency_peak_hz * sample_rate_hz)
    time_center_s = np.arange(duration_points) / sample_rate_hz
    time_center_s -= time_center_s[-1] / 2.0
    sig_gt = gt_blast_period_center(time_center_s, 1 / frequency_peak_hz)
    sig_noise = white_noise_fbits(sig_gt, noise_std_loss_bits)
    return time_center_s, antialias_half_nyquist(sig_gt + sig_noise)


def gt_blast_center_noise(
    duration_s: float = 16., frequency_peak_hz: float = 6.3,
    sample_rate_hz: float = 100., noise_std_loss_bits: float = 16.
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Fast computation of GT pulse with noise for a specified duration in seconds

    :param duration_s: signal duration in seconds.  Default 16
    :param frequency_peak_hz: peak frequency.  Default 6.3 Hz for 1 tonne TNT
    :param sample_rate_hz: sample rate.  Default 100 Hz
    :param noise_std_loss_bits: noise loss relative to signal variance.  Default 16
    :return: centered time in seconds, GT pulse with white noise
    """
    time_center_s = np.arange(int(duration_s * sample_rate_hz)) / sample_rate_hz
    time_center_s -= time_center_s[-1] / 2.0
    sig_gt = gt_blast_period_center(time_center_s, 1 / frequency_peak_hz)
    sig_noise = white_noise_fbits(sig_gt, noise_std_loss_bits)
    return time_center_s, antialias_half_nyquist(sig_gt + sig_noise)


def gt_blast_center_noise_uneven(
    sensor_epoch_s: np.array, noise_std_loss_bits: float = 2., frequency_center_hz: Optional[float] = None
) -> np.ndarray:
    """
    Construct the GT explosion pulse of Garces (2019) for even or uneven sensor time
    in Gaussian noise with SNR in bits re signal STD.
    This is a very flexible variation.

    :param sensor_epoch_s: array with timestamps for signal in epoch seconds
    :param noise_std_loss_bits: number of bits below signal standard deviation. Default is 2
    :param frequency_center_hz: Optional center frequency in Hz
    :return: numpy array with anti-aliased GT explosion pulse with Gaussian noise
    """
    time_duration_s = sensor_epoch_s[-1] - sensor_epoch_s[0]

    pseudo_period_s = 1 / frequency_center_hz if frequency_center_hz else time_duration_s / 4.0

    # Convert to seconds
    time_center_s = sensor_epoch_s - sensor_epoch_s[0] - time_duration_s / 2.0
    sig_gt = gt_blast_period_center(time_center_s, pseudo_period_s)
    sig_noise = white_noise_fbits(np.copy(sig_gt), noise_std_loss_bits)

    return antialias_half_nyquist(sig_gt + sig_noise)


# Integrals and derivatives, with delta function estimate and discontinuity boundary conditions
def gt_blast_derivative_period_center(time_center_s: np.ndarray, pseudo_period_s: float) -> np.ndarray:
    """
    Derivative of the GT blast with delta function approximation

    :param time_center_s: array with time
    :param pseudo_period_s: period in seconds
    :return: numpy ndarray with derivative of the GT blast with delta function approximation
    """
    # Garces (2019) ground truth GT blast pulse
    # with the +1, tau is the zero crossing time - time_start renamed to time_zero for first zero crossing.
    time_pos_s = pseudo_period_s / 4.0
    tau = time_center_s / time_pos_s + 1.0
    # Initialize GT
    p_gtd = np.zeros(tau.size)  # Granstrom-Triangular (GT), 2019
    # Initialize time ranges
    sigint1 = np.where((0.0 <= tau) & (tau <= 1.0))  # ONLY positive pulse
    sigint_g17 = np.where((1.0 < tau) & (tau <= 1 + np.sqrt(6.0)))  # GT balanced pulse
    p_gtd[sigint1] = -1.0
    p_gtd[sigint_g17] = -1.0 / 6.0 * (3.0 + np.sqrt(6) - 3 * tau[sigint_g17]) * (1.0 + np.sqrt(6) - tau[sigint_g17])

    return p_gtd


def gt_blast_integral_period_center(time_center_s: np.ndarray, pseudo_period_s: float) -> np.ndarray:
    """
    Integral of the GT blast with initial condition at zero crossing

    :param time_center_s: array with time
    :param pseudo_period_s: period in seconds
    :return: numpy ndarray with integral of GT blast pulse
    """
    # Garces (2019) ground truth GT blast pulse
    # with the +1, tau is the zero crossing time - time_start renamed to time_zero for first zero crossing.
    time_pos_s = pseudo_period_s / 4.0
    tau = time_center_s / time_pos_s + 1.0
    # Initialize GT
    p_gti = np.zeros(tau.size)  # Granstrom-Triangular (GT), 2019
    # Initialize time ranges
    sigint1 = np.where((0.0 <= tau) & (tau <= 1.0))  # ONLY positive pulse
    sigint_g17 = np.where((1.0 < tau) & (tau <= 1 + np.sqrt(6.0)))  # GT balanced pulse
    p_gti[sigint1] = (1.0 - tau[sigint1] / 2.0) * tau[sigint1]

    p_gti[sigint_g17] = (
        -tau[sigint_g17]
        / 72.0
        * (
            3 * tau[sigint_g17] ** 3
            - 4 * (3 + 2 * np.sqrt(6)) * tau[sigint_g17] ** 2
            + 6 * (9 + 4 * np.sqrt(6)) * tau[sigint_g17]
            - 12 * (7 + 2 * np.sqrt(6))
        )
    )

    integration_constant = p_gti[sigint1][-1] - p_gti[sigint_g17][0]
    p_gti[sigint_g17] += integration_constant

    return p_gti


def gt_blast_center_integral_and_derivative(
    frequency_peak_hz: float, sample_rate_hz: float
) -> Tuple[float, np.ndarray, np.ndarray, np.ndarray]:
    """
    Integral and derivative of GT pulse relative to tau (NOT time_s)

    :param frequency_peak_hz: peak frequency in Hz
    :param sample_rate_hz: sample rate in Hz
    :return: tau center, numpy ndarray with GT blast pulse, numpy ndarray with integral of GT blast pulse, numpy array
        with derivative of GT blast pulse
    """
    pseudo_period_s = 1 / frequency_peak_hz
    time_center_s = np.arange(int(2 / frequency_peak_hz * sample_rate_hz)) / sample_rate_hz
    time_center_s -= time_center_s[-1] / 2.0
    tau_center = time_center_s / (pseudo_period_s / 4.0)

    sig_gt = gt_blast_period_center(time_center_s, pseudo_period_s)
    sig_gt_i = gt_blast_integral_period_center(time_center_s, pseudo_period_s)
    sig_gt_d = gt_blast_derivative_period_center(time_center_s, pseudo_period_s)
    sig_gt_d[np.argmax(sig_gt) - 1] = np.max(np.diff(sig_gt)) / np.mean(np.diff(tau_center))

    return tau_center, sig_gt, sig_gt_i, sig_gt_d


def gt_blast_ft(frequency_peak_hz: float, frequency_hz: Union[float, np.ndarray]) -> Union[float, complex, np.ndarray]:
    """
    Fourier transform of the GT blast pulse

    :param frequency_peak_hz: peak frequency in Hz
    :param frequency_hz: frequency in Hz, float or np.ndarray
    :return: Fourier transform of the GT blast pulse
    """
    w_scaled = 0.5 * np.pi * frequency_hz / frequency_peak_hz
    ft_g17_positive = (1.0 - 1j * w_scaled - np.exp(-1j * w_scaled)) / w_scaled ** 2.0
    ft_g17_negative = (
        np.exp(-1j * w_scaled * (1 + np.sqrt(6.0)))
        / (3.0 * w_scaled ** 4.0)
        * (
            1j * w_scaled * np.sqrt(6.0)
            + 3.0
            + np.exp(1j * w_scaled * np.sqrt(6.0)) * (3.0 * w_scaled ** 2.0 + 1j * w_scaled * 2.0 * np.sqrt(6.0) - 3.0)
        )
    )
    return (ft_g17_positive + ft_g17_negative) * np.pi / (2 * np.pi * frequency_peak_hz)


def gt_blast_spectral_density(
    frequency_peak_hz: float, frequency_hz: Union[float, np.ndarray]
) -> Tuple[Union[float, np.ndarray], float]:
    """
    Spectral density of the GT blast pulse

    :param frequency_peak_hz: peak frequency in Hz
    :param frequency_hz: frequency in Hz, float or np.ndarray
    :return: spectral_density, spectral_density_peak
    """
    fourier_tx = gt_blast_ft(frequency_peak_hz, frequency_hz)
    spectral_density = 2 * np.abs(fourier_tx * np.conj(fourier_tx))
    return spectral_density, np.max(spectral_density)



import datetime
import unittest

import pytz

from quantum_inferno.utilities import date_time as dat


class TimeUnitTest(unittest.TestCase):
    def test_picos_to_seconds(self):  # use large number to avoid floating point error
        picos = 1e12 * dat.time_unit_dict["ps"]
        self.assertEqual(picos, 1.0)

    def test_nanos_to_seconds(self):  # use large number to avoid floating point error
        nanos = 1e9 * dat.time_unit_dict["ns"]
        self.assertEqual(nanos, 1.0)

    def test_micros_to_seconds(self):  # use large number to avoid floating point error
        micros = 1e6 * dat.time_unit_dict["us"]
        self.assertEqual(micros, 1.0)

    def test_millis_to_seconds(self):
        millis = 100.0 * dat.time_unit_dict["ms"]
        self.assertEqual(millis, 0.1)

    def test_seconds_to_seconds(self):
        seconds = 100.0 * dat.time_unit_dict["s"]
        self.assertEqual(seconds, 100)

    def test_minutes_to_seconds(self):
        minutes = 100.0 * dat.time_unit_dict["m"]
        self.assertEqual(minutes, 6000)

    def test_hours_to_seconds(self):
        hours = 100.0 * dat.time_unit_dict["h"]
        self.assertEqual(hours, 360000)

    def test_days_to_seconds(self):
        days = 100.0 * dat.time_unit_dict["d"]
        self.assertEqual(days, 8640000)

    def test_weeks_to_seconds(self):
        weeks = 100.0 * dat.time_unit_dict["weeks"]
        self.assertEqual(weeks, 60480000)

    def test_months_to_seconds(self):
        months = 100.0 * dat.time_unit_dict["months"]
        self.assertEqual(months, 262800000)

    def test_years_to_seconds(self):
        years = 100.0 * dat.time_unit_dict["years"]
        self.assertEqual(years, 3153600000)


class ConvertTimeTest(unittest.TestCase):
    def test_convert_time_to_seconds(self):
        time_seconds = dat.convert_time_unit(100.0, "m", "s")
        self.assertEqual(time_seconds, 6000)

    def test_fail_convert(self):
        self.assertRaises(ValueError, dat.convert_time_unit, 0, "x", "y")


class DatetimeToTimestampTest(unittest.TestCase):
    def test_datetime_to_timestamp(self):
        datetime_obj = dat.datetime(2021, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)
        timestamp = dat.utc_datetime_to_utc_timestamp(datetime_obj, "s")
        self.assertEqual(timestamp, 1609459200.0)

    def test_datetime_to_timestamp_no_tz(self):
        datetime_obj = dat.datetime(2021, 1, 1, 0, 0, 0)
        timestamp = dat.utc_datetime_to_utc_timestamp(datetime_obj, "s")
        self.assertEqual(timestamp, 1609459200.0)

    def test_fail_timestamp(self):
        self.assertRaises(ValueError, dat.utc_datetime_to_utc_timestamp, 0, "x")


class TimestampToDatetimeTest(unittest.TestCase):
    def test_timestamp_to_datetime(self):
        datetime_obj = dat.utc_timestamp_to_utc_datetime(1609459200.0, "s")
        self.assertEqual(datetime_obj, dat.datetime(2021, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc))

    def test_fail_datetime(self):
        self.assertRaises(ValueError, dat.utc_timestamp_to_utc_datetime, 0, "x")


class SetDatetimeToUtcTest(unittest.TestCase):
    def test_set_datetime_to_utc(self):
        datetime_obj = dat.datetime(2021, 1, 1, 0, 0, 0, tzinfo=pytz.timezone("HST"))
        utc_datetime_obj = dat.set_datetime_to_utc(datetime_obj)
        self.assertEqual(utc_datetime_obj, dat.datetime(2021, 1, 1, 10, 0, 0, tzinfo=datetime.timezone.utc))

    def test_set_datetime_to_utc_warning(self):
        datetime_obj = dat.datetime(2021, 1, 1, 0, 0, 0)
        utc_datetime_obj = dat.set_datetime_to_utc(datetime_obj, tzinfo_warning=False)
        self.assertEqual(utc_datetime_obj, dat.datetime(2021, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc))


class SetTimestampToUtcTest(unittest.TestCase):
    def test_set_timestamp_to_utc(self):
        utc_timestamp = dat.set_timestamp_to_utc(1609462800.0, 1, "s")
        self.assertEqual(utc_timestamp, 1609459200)

        utc_timestamp = dat.set_timestamp_to_utc(1609455600.0, -1, "s")
        self.assertEqual(utc_timestamp, 1609459200)

    def test_fail_timestamp(self):
        self.assertRaises(ValueError, dat.set_timestamp_to_utc, 0, 0, "x")


class GetDatetimeFromTimestampToUtcTest(unittest.TestCase):
    def test_get_datetime_from_timestamp_to_utc(self):
        datetime_obj = dat.get_datetime_from_timestamp_to_utc(1609459200.0, -1.0, "s")
        # print(datetime_obj)
        self.assertEqual(datetime_obj, dat.datetime(2021, 1, 1, 1, 0, 0, tzinfo=datetime.timezone.utc))

    def test_fail_get_datetime(self):
        self.assertRaises(ValueError, dat.get_datetime_from_timestamp_to_utc, 1609459200.0, -1.0, "x")



import unittest

import numpy as np

import quantum_inferno.synth.benchmark_signals as bs


class TestBenchmarkSignals(unittest.TestCase):
    def test_signal_gate(self):
        self.assertEqual(1, 1)



"""
This module contains quick plotting routines
"""
import numpy as np
import matplotlib.pyplot as plt

from quantum_inferno.scales_dyadic import get_epsilon


def plot_tdr_sig(sig_wfm: np.ndarray, sig_time: np.ndarray, signal_time_base: str = "seconds") -> plt.Figure:
    """
    Waveform

    :param sig_wfm:
    :param sig_time:
    :param signal_time_base:
    :return: figure of waveform for plotting
    """
    fig = plt.figure()
    plt.plot(sig_time, sig_wfm)
    plt.title("Input waveform")
    plt.xlabel("Time, " + signal_time_base)
    return fig


def plot_tdr_rms(
        sig_wfm: np.ndarray,
        sig_time: np.ndarray,
        sig_rms_wf: np.ndarray,
        sig_rms_time: np.ndarray,
        signal_time_base: str = "seconds"
) -> plt.Figure:
    """
    Waveform

    :param sig_wfm:
    :param sig_time:
    :param sig_rms_wf:
    :param sig_rms_time:
    :param signal_time_base:
    :return: figure of waveform for plotting
    """
    fig = plt.figure()
    plt.plot(sig_time, sig_wfm)
    plt.plot(sig_rms_time, sig_rms_wf)
    plt.title("Input waveform and RMS")
    plt.xlabel("Time, " + signal_time_base)
    return fig


def plot_tfr_lin(
        tfr_power: np.ndarray,
        tfr_frequency: np.ndarray,
        tfr_time: np.ndarray,
        title_str: str = "TFR, power",
        signal_time_base: str = "seconds"
) -> plt.Figure:
    """
    TFR in linear power

    :param tfr_power:
    :param tfr_frequency:
    :param tfr_time:
    :param title_str:
    :param signal_time_base:
    :return: figure of waveform for plotting
    """
    fig = plt.figure()
    plt.pcolormesh(tfr_time, tfr_frequency, tfr_power, cmap="RdBu_r")
    plt.title(title_str)
    plt.ylabel("Frequency, samples per " + signal_time_base)
    plt.xlabel("Time, " + signal_time_base)
    return fig


def plot_tfr_bits(
        tfr_power: np.ndarray,
        tfr_frequency: np.ndarray,
        tfr_time: np.ndarray,
        bits_min: float = -8,
        bits_max: float = 0,
        title_str: str = "TFR, top bits",
        y_scale: str = None,
        tfr_x_str: str = "Time, seconds",
        tfr_y_str: str = "Frequency, hz",
        tfr_y_flip: bool = False,
) -> plt.Figure:
    """
    TFR in bits

    :param tfr_power:
    :param tfr_frequency:
    :param tfr_time:
    :param bits_max:
    :param bits_min:
    :param y_scale:
    :param title_str:
    :param tfr_x_str:
    :param tfr_y_str:
    :param tfr_y_flip:
    :return: figure of waveform for plotting
    """
    tfr_bits = 0.5 * np.log2(tfr_power / np.max(tfr_power))

    fig = plt.figure()
    plt.pcolormesh(tfr_time, tfr_frequency, tfr_bits, cmap="RdBu_r", vmin=bits_min, vmax=bits_max, shading="nearest")
    plt.yscale("linear" if y_scale is None else "log")

    if tfr_y_flip:
        plt.ylim(np.max(tfr_frequency), np.min(tfr_frequency))
    plt.title(title_str)
    plt.ylabel(tfr_y_str)
    plt.xlabel(tfr_x_str)

    return fig


def plot_st_window_tdr_lin(window: np.ndarray, freq_sx: np.ndarray, time_fft: np.ndarray) -> plt.Figure:
    """
    plot something

    :param window:
    :param freq_sx:
    :param time_fft:
    :return: figure of waveform for plotting
    """
    fig = plt.figure(figsize=(8, 8))
    for j, freq in enumerate(freq_sx):
        plt.plot(time_fft, np.abs(window[j, :]), label=freq)
    plt.legend()
    plt.title("TDR window, linear")
    return fig


def plot_st_window_tfr_bits(window: np.ndarray, frequency_sx: np.ndarray, frequency_fft: np.ndarray) -> plt.Figure:
    """
    plot something

    :param window:
    :param frequency_sx:
    :param frequency_fft:
    :return: figure of waveform for plotting
    """
    fig = plt.figure(figsize=(8, 8))
    for j, freq in enumerate(frequency_sx):
        plt.plot(frequency_fft, np.log2(np.abs(window[j, :]) + get_epsilon()), label=freq)
    plt.legend()
    plt.title("TFR window, bits")
    return fig


def plot_st_window_tfr_lin(window: np.ndarray, frequency_sx: np.ndarray, frequency_fft: np.ndarray) -> plt.Figure:
    """
    plot something

    :param window:
    :param frequency_sx:
    :param frequency_fft:
    :return: figure of waveform for plotting
    """
    fig = plt.figure(figsize=(8, 8))
    for j, freq in enumerate(frequency_sx):
        plt.plot(frequency_fft, np.abs(window[j, :]), label=freq)
    plt.legend()
    plt.title("TFR window, lin")
    return fig



"""
This module returns information and entropy from a TFR Power array
"""
from typing import Tuple

import numpy as np
import scipy.fft as sfft

import quantum_inferno.scales_dyadic as scales
from quantum_inferno.utilities.matrix import d1tile_x_d0d1, d0tile_x_d0d1


def log2_ceil(x: float, epsilon: float = scales.EPSILON64) -> float:
    """
    Compute ceiling of log2 of a positive input argument.
    Corrects for negative, complex or zero inputs by taking the absolute value and adding EPSILON

    :param x: input, converts to positive real
    :param epsilon: override zero, negative, and imaginary values
    :return: ceiling of log2
    """
    return np.ceil(np.log2(np.abs(x) + epsilon))


def log2_round(x: float, epsilon: float = scales.EPSILON64) -> float:
    """
    Compute rounded value of log2 of a positive input argument.
    Corrects for negative, complex or zero inputs by taking the absolute value and adding EPSILON

    :param x: input, converts to positive real
    :param epsilon: override zero, negative, and imaginary values
    :return: rounded to nearest integer value of log2
    """
    return float(np.round(np.log2(np.abs(x) + epsilon)))


def log2_floor(x: float, epsilon: float = scales.EPSILON64) -> float:
    """
    Compute floor of log2 of a positive input argument.
    Corrects for negative, complex or zero inputs by taking the absolute value and adding EPSILON

    :param x: input, converts to positive real
    :param epsilon: override zero, negative, and imaginary values
    :return: floor of log2
    """
    return np.floor(np.log2(np.abs(x) + epsilon))


def mat_max_idx(a: np.ndarray) -> Tuple[np.ndarray]:
    """
    :param a: matrix to find maximum for
    :return: The indexes of the max of a matrix
    """
    return np.unravel_index(a.argmax(), a.shape)


def mat_min_idx(a: np.ndarray) -> Tuple[np.ndarray]:
    """
    :param a: matrix to find minimum for
    :return: The indexes of the min of a matrix
    """
    return np.unravel_index(a.argmin(), a.shape)


def scale_log2_64(in_array: np.ndarray) -> np.ndarray:
    """
    :param in_array: input array
    :return: log2 of array values plus EPSILON64
    """
    return np.log2(in_array + scales.EPSILON64)


def scale_power_bits(power: np.ndarray) -> np.ndarray:
    """
    :param power: power from time-frequency representation
    :return: scaled power bits minus the maximum value
    """
    power_bits = scale_log2_64(power)
    return power_bits - np.max(power_bits)


def power_dynamics_scaled_bits(tfr_power: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Essential scales for power

    :param tfr_power: power from time-frequency representation
    :return: Log2 scaled power bits, dynamic time range, dynamic frequency range
    """
    tfr_power_bits = scale_power_bits(tfr_power)  # Log2 scaled power bits
    # Dynamic range per time step re max
    tfr_power_per_time_bits = scale_power_bits(np.sum(tfr_power, axis=0))
    # Dynamic range per frequency band re max
    tfr_power_per_freq_bits = scale_power_bits(np.sum(tfr_power, axis=1))
    return tfr_power_bits, tfr_power_per_time_bits, tfr_power_per_freq_bits


def get_info_and_entropy_32(marginal: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    :param marginal: array of values to start with
    :return: info, entropy and ref_entropy of the array, using EPSILON32
    """
    info = -np.log2(marginal + scales.EPSILON32)
    return info, marginal * info, np.log2(len(marginal)) / len(marginal)


class Shannon:
    """
    class for Shannon information of a signal

    Attributes:
        marginal: np.ndarray, waveform

        info: np.ndarray, info of waveform

        entropy: np.ndarray, entropy of waveform

        ref_entropy: np.ndarray, reference entropy of waveform

        isnr: np.ndarray, isnr of waveform

        esnr: np.ndarray, esnr of waveform
    """

    def __init__(self, marginal: np.ndarray):
        """
        :param marginal: waveform
        """
        self.marginal: np.ndarray = marginal
        self.info: np.ndarray
        self.entropy: np.ndarray
        self.ref_entropy: np.ndarray
        # Average entropy for P ~ 1 / NFFT
        self.info, self.entropy, self.ref_entropy = get_info_and_entropy_32(self.marginal)
        self.isnr: np.ndarray = np.log2(len(self.info)) - self.info
        self.esnr: np.ndarray = self.entropy / self.ref_entropy


class ShannonTDR(Shannon):
    """
    class for Shannon TDR information, refer to Shannon for base class

    Attributes:
        sig: np.ndarray, data to process
    """
    def __init__(self, sig_in_real: np.ndarray):
        """
        :param sig_in_real: data to process
        """
        self.sig: np.ndarray = sig_in_real / np.sqrt(np.sum(sig_in_real ** 2))
        super().__init__(self.sig ** 2)

    def print_total_ref_entropy(self):
        print("Ref entropy, time:", self.ref_entropy)

    def print_total_entropy(self):
        print("Total Entropy, time:", np.sum(self.entropy))

    def print_total_marginal(self):
        print("Sum of time marginal:", np.sum(self.marginal))


class ShannonFFT(Shannon):
    """
    class for Shannon FFT information, refer to Shannon for base class

    Attributes:
        sig: np.ndarray, data to process

        angle_rads: np.ndarray, angle in radians

        frequency: np.ndarray, frequencies of the data
    """
    def __init__(self, sig_in_real: np.ndarray):
        """
        :param sig_in_real: data to process
        """
        self.sig: np.ndarray = sfft.rfft(x=sig_in_real)
        self.angle_rads: np.ndarray = np.unwrap(np.angle(self.sig))
        self.frequency: np.ndarray = np.arange(len(self.angle_rads)) / len(self.angle_rads) / 2.0
        fft_sq = np.abs(self.sig) ** 2
        super().__init__(fft_sq / np.sum(fft_sq))

    def print_total_ref_entropy(self):
        print("Ref entropy, frequency:", self.ref_entropy)

    def print_total_entropy(self):
        print("Total Entropy, frequency:", np.sum(self.entropy))

    def print_total_marginal(self):
        print("Sum of frequency marginal:", np.sum(self.marginal))


def shannon_tdr_fft(sig_in_real: np.ndarray) -> Tuple[ShannonTDR, ShannonFFT]:
    """
    Shannon information and entropy

    :param sig_in_real: data to process
    :return: ShannonTDR and ShannonFFT of the data
    """
    return ShannonTDR(sig_in_real), ShannonFFT(sig_in_real)


class ShannonStft:
    """
    Class for Shannon stft

    Attributes:
        info: np.ndarray, info of waveform

        shannon_bits: np.ndarray, shannon bits of waveform

        ref_bits: np.ndarray, ref bits of waveform

        isnr: np.ndarray, isnr of waveform

        esnr: np.ndarray, esnr of waveform
    """

    def __init__(self, tfr_pow_pdf: np.ndarray, deg_free: int):
        """
        :param tfr_pow_pdf: tfr power matrix
        :param deg_free: degrees of freedom
        """
        self.info: np.ndarray = -scale_log2_64(tfr_pow_pdf)
        self.shannon_bits: np.ndarray = tfr_pow_pdf * self.info
        self.ref_bits: float = np.log2(deg_free) / deg_free
        self.isnr: np.ndarray = np.log2(deg_free) - self.info
        self.esnr: np.ndarray = self.shannon_bits / self.ref_bits


def shannon_stft_from_tfr_power(tfr_power: np.ndarray) -> ShannonStft:
    """
    :param tfr_power: tfr power matrix
    :return: ShannonStft of the data
    """
    return ShannonStft(tfr_power / np.sum(tfr_power), tfr_power.shape[0] * tfr_power.shape[1])


class ShannonStftPerTime(ShannonStft):
    """
    Class for Shannon stft per time, refer to ShannonStft for base class
    """
    def __init__(self, tfr_power: np.ndarray):
        """
        :param tfr_power: tfr power matrix
        """
        tfr_power_per_time_pdf = d1tile_x_d0d1(d1=1 / np.sum(tfr_power, axis=0) + scales.EPSILON64, d0d1=tfr_power)
        super().__init__(tfr_power_per_time_pdf, tfr_power.shape[0])


class ShannonStftPerFreq(ShannonStft):
    """
    Class for Shannon stft per frequency, refer to ShannonStft for base class
    """
    def __init__(self, tfr_power: np.ndarray):
        """
        :param tfr_power: tfr power matrix
        """
        tfr_power_per_freq_pdf = d0tile_x_d0d1(d0=1 / np.sum(tfr_power, axis=1) + scales.EPSILON64, d0d1=tfr_power)
        super().__init__(tfr_power_per_freq_pdf, tfr_power.shape[1])



import unittest

from quantum_inferno.utilities import rescaling


class MyTestCase(unittest.TestCase):
    def test_to_log2_with_epsilon(self):
        result = rescaling.to_log2_with_epsilon(100.0)
        self.assertAlmostEqual(result, 6.64, 2)
        result = rescaling.to_log2_with_epsilon(-100.0)
        self.assertAlmostEqual(result, 6.64, 2)
        values = [100.0, -100.0]
        result = rescaling.to_log2_with_epsilon(values)
        self.assertAlmostEqual(result[0], 6.64, 2)
        self.assertAlmostEqual(result[1], 6.64, 2)

    def test_is_power_of_two(self):
        result = rescaling.is_power_of_two(8)
        self.assertTrue(result)
        result = rescaling.is_power_of_two(9)
        self.assertFalse(result)

    def test_to_decibel_with_epsilon(self):
        result = rescaling.to_decibel_with_epsilon(100.0, 1.0)
        self.assertEqual(result, 40.0)
        result = rescaling.to_decibel_with_epsilon(-100.0, 1.0)
        self.assertEqual(result, 40.0)
        result = rescaling.to_decibel_with_epsilon(100.0, 1.0, "amplitude")
        self.assertEqual(result, 40.0)
        result = rescaling.to_decibel_with_epsilon(100.0, 1.0, "power")
        self.assertEqual(result, 20.0)
        result = rescaling.to_decibel_with_epsilon(100.0, 100.0)
        self.assertEqual(result, 0)
        values = [100.0, -100.0]
        result = rescaling.to_decibel_with_epsilon(values, 1.0)
        self.assertEqual(result[0], 40.0)
        self.assertEqual(result[1], 40.0)



"""
This module calculates spectra: STFT, FFT
"""
from typing import Tuple, Union

import numpy as np
import scipy.signal as signal

from quantum_inferno import scales_dyadic as scales


def butter_bandpass(
    sig_wf: np.ndarray,
    frequency_sample_rate_hz: float,
    frequency_cut_low_hz,
    frequency_cut_high_hz,
    filter_order: int = 4,
    tukey_alpha: float = 0.5,
) -> np.ndarray:
    """
    Butterworth bandpass filter

    :param sig_wf: signal waveform as numpy array
    :param frequency_sample_rate_hz: frequency sample rate in Hz
    :param frequency_cut_low_hz: frequency low value
    :param frequency_cut_high_hz: frequency high value
    :param filter_order: filter order
    :param tukey_alpha: Tukey window alpha
    :return: filtered signal waveform as numpy array
    """
    nyquist = 0.5 * frequency_sample_rate_hz
    edge_low = frequency_cut_low_hz / nyquist
    edge_high = frequency_cut_high_hz / nyquist
    if edge_high >= 1:
        print(
            f"Warning: Frequency cutoff {frequency_cut_high_hz} greater than Nyquist {nyquist} Hz, using half Nyquist"
        )
        edge_high = 0.5  # Half of nyquist
    [b, a] = signal.butter(N=filter_order, Wn=[edge_low, edge_high], btype="bandpass")
    sig_taper = np.copy(sig_wf)
    sig_taper *= signal.windows.tukey(M=len(sig_taper), alpha=tukey_alpha)
    return signal.filtfilt(b, a, sig_taper)


def butter_highpass(
    sig_wf: np.ndarray,
    frequency_sample_rate_hz: float,
    frequency_cut_low_hz: Union[float, int],
    filter_order: int = 4,
    tukey_alpha: float = 0.5,
) -> np.ndarray:
    """
    Butterworth bandpass filter

    :param sig_wf: signal waveform as numpy array
    :param frequency_sample_rate_hz: frequency sample rate in Hz
    :param frequency_cut_low_hz: frequency low value
    :param filter_order: filter order
    :param tukey_alpha: Tukey window alpha
    :return: filtered signal waveform as numpy array
    """
    edge_low = frequency_cut_low_hz / (0.5 * frequency_sample_rate_hz)

    if edge_low >= 1:
        raise ValueError(
            f"Frequency cutoff {frequency_cut_low_hz} is greater than Nyquist {0.5*frequency_sample_rate_hz}"
        )

    [b, a] = signal.butter(N=filter_order, Wn=[edge_low], btype="highpass")
    sig_taper = np.copy(sig_wf)
    sig_taper *= signal.windows.tukey(M=len(sig_taper), alpha=tukey_alpha)
    return signal.filtfilt(b, a, sig_taper)


def butter_lowpass(
    sig_wf: np.ndarray,
    frequency_sample_rate_hz: float,
    frequency_cut_high_hz: Union[float, int],
    filter_order: int = 4,
    tukey_alpha: float = 0.5,
) -> np.ndarray:
    """
    Butterworth bandpass filter

    :param sig_wf: signal waveform as numpy array
    :param frequency_sample_rate_hz: frequency sample rate in Hz
    :param frequency_cut_high_hz: frequency low value
    :param filter_order: filter order
    :param tukey_alpha: Tukey window alpha
    :return: filtered signal waveform as numpy array
    """
    edge_high = frequency_cut_high_hz / (0.5 * frequency_sample_rate_hz)

    if edge_high >= 1:
        raise ValueError(
            f"Frequency cutoff {frequency_cut_high_hz} is greater than Nyquist {0.5*frequency_sample_rate_hz}"
        )
    [b, a] = signal.butter(N=filter_order, Wn=[edge_high], btype="lowpass")
    sig_taper = np.copy(sig_wf)
    sig_taper *= signal.windows.tukey(M=len(sig_taper), alpha=tukey_alpha)
    return signal.filtfilt(b, a, sig_taper)


def stft_complex_pow2(
    sig_wf: np.ndarray,
    frequency_sample_rate_hz: float,
    segment_points: int,
    overlap_points: int = None,
    nfft_points: int = None,
    alpha: float = 0.25,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Simplest, with 50% overlap and built-in defaults

    :param sig_wf: signal waveform as numpy array
    :param frequency_sample_rate_hz: frequency sample rate in Hz
    :param segment_points: number of points in a segment
    :param overlap_points: number of points in overlap
    :param nfft_points: length of FFT
    :param alpha: Tukey window alpha
    :return: frequency_stft_hz, time_stft_s, stft_complex
    """
    if nfft_points is None:
        nfft_points = int(2 ** np.ceil(np.log2(segment_points)))
    if overlap_points is None:
        overlap_points = int(segment_points / 2)
    return signal.stft(
        x=sig_wf,
        fs=frequency_sample_rate_hz,
        window=("tukey", alpha),
        nperseg=segment_points,
        noverlap=overlap_points,
        nfft=nfft_points,
        detrend="constant",
        return_onesided=True,
        axis=-1,
        boundary="zeros",
        padded=True,
    )


def gtx_complex_pow2(
    sig_wf: np.ndarray,
    frequency_sample_rate_hz: float,
    segment_points: int,
    gaussian_sigma: int = None,
    overlap_points: int = None,
    nfft_points: int = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Gaussian taper with 50% overlap and built-in defaults

    :param sig_wf: signal waveform as numpy array
    :param frequency_sample_rate_hz: frequency sample rate in Hz
    :param segment_points: number of points in a segment
    :param gaussian_sigma: gaussian window variance
    :param overlap_points: number of points in overlap
    :param nfft_points: length of FFT
    :return: frequency_stft_hz, time_stft_s, stft_complex
    """
    if nfft_points is None:
        nfft_points = int(2 ** np.ceil(np.log2(segment_points)))
    if overlap_points is None:
        overlap_points = int(segment_points / 2)
    if gaussian_sigma is None:
        gaussian_sigma = int(segment_points / 4)
    return signal.stft(
        x=sig_wf,
        fs=frequency_sample_rate_hz,
        window=("gaussian", gaussian_sigma),
        nperseg=segment_points,
        noverlap=overlap_points,
        nfft=nfft_points,
        detrend="constant",
        return_onesided=True,
        axis=-1,
        boundary="zeros",
        padded=True,
    )


def welch_power_pow2(
    sig_wf: np.ndarray,
    frequency_sample_rate_hz: float,
    segment_points: int,
    nfft_points: int = None,
    overlap_points: int = None,
    alpha: float = 0.25,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Simplest, with 50% overlap and built-in defaults

    :param sig_wf: signal waveform as numpy array
    :param frequency_sample_rate_hz: frequency sample rate in Hz
    :param segment_points: number of points in a segment
    :param overlap_points: number of points in overlap
    :param nfft_points: length of FFT
    :param alpha: Tukey window alpha
    :return: frequency_welch_hz, welch_power
    """
    if nfft_points is None:
        nfft_points = int(2 ** np.ceil(np.log2(segment_points)))
    if overlap_points is None:
        overlap_points = int(segment_points / 2)
    # Compute the Welch PSD; averaged spectrum over sliding windows
    return signal.welch(
        x=sig_wf,
        fs=frequency_sample_rate_hz,
        window=("tukey", alpha),
        nperseg=segment_points,
        noverlap=overlap_points,
        nfft=nfft_points,
        detrend="constant",
        return_onesided=True,
        axis=-1,
        scaling="spectrum",
        average="mean",
    )



import unittest

import quantum_inferno.plot_templates.figure_attributes as fa


class FigureAttributesTest(unittest.TestCase):
    def test_figure_attributes(self):
        test_fa = fa.FigureAttributes()
        self.assertEqual(test_fa.fig_scale, 2.0)
        self.assertEqual(test_fa.fig_dpi, 300)
        self.assertEqual(test_fa.ratio[0], 640)
        self.assertEqual(test_fa.ratio[1], 400)
        self.assertEqual(test_fa.font_size_1st_level, 10)
        self.assertEqual(test_fa.font_size_2nd_level, 8)
        self.assertEqual(test_fa.line_color, "k")
        self.assertEqual(test_fa.line_style, "-")
        self.assertEqual(test_fa.fig_aspect_ratio[0], 1280)
        self.assertEqual(test_fa.fig_aspect_ratio[1], 800)
        self.assertEqual(test_fa.fig_face_color, "w")
        self.assertEqual(test_fa.fig_edge_color, "w")
        self.assertAlmostEqual(test_fa.fig_size[0], 4.267, 2)
        self.assertAlmostEqual(test_fa.fig_size[1], 2.67, 2)
        self.assertEqual(test_fa.font_color, "k")
        self.assertEqual(test_fa.font_weight, "normal")
        self.assertEqual(test_fa.line_weight, 2)
        self.assertEqual(test_fa.tick_size, 8)
        self.assertEqual(test_fa.legend_label_size, 8)
        self.assertIsNone(test_fa.fig)


class FigureAttributesBlackTest(unittest.TestCase):
    def test_fa_black(self):
        test_fa = fa.FigureAttributesBackInBlack()
        self.assertEqual(test_fa.fig_face_color, "k")
        self.assertEqual(test_fa.fig_edge_color, "k")
        self.assertEqual(test_fa.font_color, "w")


class FigureParametersTest(unittest.TestCase):
    def test_figure_params_640_360(self):
        test_fp = fa.FigureParameters(fa.AspectRatioType.R640x360)
        self.assertEqual(test_fp.width, 640)
        self.assertEqual(test_fp.height, 360)
        self.assertAlmostEqual(test_fp.scale_factor, .33, 2)
        self.assertEqual(test_fp.figure_size_x, 42)
        self.assertEqual(test_fp.figure_size_y, 24)
        self.assertAlmostEqual(test_fp.text_size, 48)

    def test_figure_params_1280_720(self):
        test_fp = fa.FigureParameters(fa.AspectRatioType.R1280x720)
        self.assertEqual(test_fp.width, 1280)
        self.assertEqual(test_fp.height, 720)
        self.assertAlmostEqual(test_fp.scale_factor, .666, 2)
        self.assertEqual(test_fp.figure_size_x, 21)
        self.assertEqual(test_fp.figure_size_y, 12)
        self.assertAlmostEqual(test_fp.text_size, 24)

    def test_figure_params_1920_1080(self):
        test_fp = fa.FigureParameters(fa.AspectRatioType.R1920x1080)
        self.assertEqual(test_fp.width, 1920)
        self.assertEqual(test_fp.height, 1080)
        self.assertEqual(test_fp.scale_factor, 1.25)
        self.assertEqual(test_fp.figure_size_x, 11)
        self.assertEqual(test_fp.figure_size_y, 6)
        self.assertAlmostEqual(test_fp.text_size, 12)

    def test_figure_params_2560_1440(self):
        test_fp = fa.FigureParameters(fa.AspectRatioType.R2560x1440)
        self.assertEqual(test_fp.width, 2560)
        self.assertEqual(test_fp.height, 1440)
        self.assertAlmostEqual(test_fp.scale_factor, 1.33, 2)
        self.assertEqual(test_fp.figure_size_x, 10)
        self.assertEqual(test_fp.figure_size_y, 6)
        self.assertAlmostEqual(test_fp.text_size, 12)

    def test_figure_params_3840_2160(self):
        test_fp = fa.FigureParameters(fa.AspectRatioType.R3840x2160)
        self.assertEqual(test_fp.width, 3840)
        self.assertEqual(test_fp.height, 2160)
        self.assertEqual(test_fp.scale_factor, 2)
        self.assertEqual(test_fp.figure_size_x, 7)
        self.assertEqual(test_fp.figure_size_y, 4)
        self.assertAlmostEqual(test_fp.text_size, 8)


class AudioParamsTest(unittest.TestCase):
    def test_audio_params(self):
        test_ap = fa.AudioParams()
        self.assertEqual(test_ap.width, 1920)
        self.assertEqual(test_ap.height, 1080)
        self.assertEqual(test_ap.scale_factor, 1.25)
        self.assertEqual(test_ap.figure_size_x, 11)
        self.assertEqual(test_ap.figure_size_y, 6)
        self.assertAlmostEqual(test_ap.text_size, 12)
        self.assertTrue(test_ap.fill_gaps)



"""
This module contains functions to construct quantized, standardized information packets using binary metrics.
No-chirp/sweep (index_shift=0, variable removed), simplified for the base stockwell transform.
Based on Garces (2023).
"""
from typing import Tuple, Union

import numpy as np
import scipy.signal as signal

from quantum_inferno import scales_dyadic as scales


def wavelet_variance_theory(amp: float, time_s: np.ndarray, scale: float, omega: float) -> Tuple[float, float]:
    """
    Theoretical variance of a Gabor wavelet, with real and imaginary components

    :param amp: wavelet amplitude
    :param time_s: time support vector
    :param scale: wavelet scale
    :param omega: angular frequency
    :return: nominal wavelet real and imaginary variance
    """
    base_var = amp**2 / len(time_s) * 0.5 * np.sqrt(np.pi) * scale
    return base_var / (1 + np.exp(-(scale * omega)**2)), base_var / (1 - np.exp(-(scale * omega)**2))


def wavelet_amplitude(scale_atom: Union[np.ndarray, float]) -> \
        Tuple[Union[np.ndarray, float], Union[np.ndarray, float]]:
    """
    :param scale_atom: atom/logon scale
    :return: amp_canonical, amp_unit_spectrum
    """
    # amp_canonical = return unit integrated power and spectral energy. Good for math, ref William et al. 1991.
    # amp_unit_spectrum = return unit peak spectrum; for practical implementation.
    # Programmers: Although tempting, do not simplify - this follows the original math and is a touchstone.
    amp_canonical = (np.pi * scale_atom ** 2) ** (-1/4)
    amp_unit_spectrum = (4 * np.pi * scale_atom ** 2) ** (-1/4) * amp_canonical
    return amp_canonical, amp_unit_spectrum


def amplitude_convert_norm_to_spect(scale_atom: Union[np.ndarray, float]) -> \
        Tuple[Union[np.ndarray, float], Union[np.ndarray, float]]:
    """
    :param scale_atom: atom/logon scale
    :return: amp_canonical, amp_unit_spectrum
    """
    # amp_canonical = return unit integrated power and spectral energy. Good for math, ref William et al. 1991.
    # amp_unit_spectrum = return unit peak spectrum; for practical implementation.
    amp_canonical = (np.pi * scale_atom ** 2) ** (-1/4)
    amp_unit_spectrum = (4 * np.pi * scale_atom**2) ** (-1/4) * amp_canonical
    amp_norm2spect = amp_unit_spectrum/amp_canonical

    return amp_norm2spect


def wavelet_time(time_s: np.ndarray, offset_time_s: float, frequency_sample_rate_hz: float) -> np.ndarray:
    """
    :param time_s: array with time
    :param offset_time_s: offset time in seconds
    :param frequency_sample_rate_hz: sample rate in Hz
    :return: numpy array with scaled time-shifted time
    """
    return frequency_sample_rate_hz * (time_s - offset_time_s)


def wavelet_complex(
        band_order_nth: float,
        time_s: np.ndarray,
        offset_time_s: float,
        scale_frequency_center_hz: Union[np.ndarray, float],
        frequency_sample_rate_hz: float
) -> Tuple[np.ndarray, np.ndarray, Union[np.ndarray, float], Union[np.ndarray, float],
           Union[np.ndarray, float], Union[np.ndarray, float], Union[np.ndarray, float]]:
    """
    Quantized atom for specified band_order_Nth and arbitrary time duration.
    Returns a frequency x time dimension wavelet vector

    :param band_order_nth: Nth order of constant Q bands
    :param time_s: time in seconds, duration should be greater than or equal to M/fc
    :param offset_time_s: offset time in seconds, should be between min and max of time_s
    :param scale_frequency_center_hz: center frequency fc in Hz
    :param frequency_sample_rate_hz: sample rate on Hz
    :return: waveform_complex, time_shifted_s, scale_angular_frequency, scale, omega, amp_canonical, amp_unit_spectrum
    """
    # Center and nondimensionalize time
    xtime_shifted = wavelet_time(time_s, offset_time_s, frequency_sample_rate_hz)

    # Nondimensional chirp parameters
    scale_atom, scale_angular_frequency = \
        scales.scale_from_frequency_hz(band_order_nth, scale_frequency_center_hz, frequency_sample_rate_hz)

    if np.isscalar(scale_atom):
        # Single frequency input
        xtime = xtime_shifted
        scale = scale_atom
        omega = scale_angular_frequency
    else:
        # Convert scale, frequency and time vectors to [frequency x time] matrices
        xtime = np.tile(xtime_shifted, (len(scale_atom), 1))
        scale = np.tile(scale_atom, (len(xtime_shifted), 1)).T
        omega = np.tile(scale_angular_frequency, (len(xtime_shifted), 1)).T

    # Base wavelet with unit absolute amplitude.
    # Note centered imaginary wavelet (sine) does not reach unity because of Gaussian envelope.
    wavelet_gabor = np.exp(-0.5*(xtime/scale)**2) * np.exp(1j*omega*xtime)
    amp_canonical, amp_unit_spectrum = wavelet_amplitude(scale)

    return wavelet_gabor, xtime_shifted, scale_angular_frequency, scale, omega, amp_canonical, amp_unit_spectrum


def wavelet_centered_4cwt(
        band_order_nth: float,
        duration_points: int,
        scale_frequency_center_hz: Union[np.ndarray, float],
        frequency_sample_rate_hz: float,
        dictionary_type: str = "norm"
) -> Tuple[np.ndarray, np.ndarray, Union[np.ndarray, float], Union[np.ndarray, float], Union[np.ndarray, float]]:
    """
    Gabor atoms for CWT computation centered on the duration of signal

    Options for dictionary_type: "norm" (Canonical unit-norm), "spect" (unit spectrum), or "unit" (unit modulus)

    :param duration_points: number of points in the signal
    :param band_order_nth: Nth order of constant Q bands
    :param scale_frequency_center_hz: center frequency fc in Hz
    :param frequency_sample_rate_hz: sample rate is Hz
    :param dictionary_type: determines amplification value.  Default "norm"
    :return: waveform_complex, time_shifted_s, scale, omega, amp
    """
    time_s = np.arange(duration_points) / frequency_sample_rate_hz

    wavelet_gabor, xtime_shifted, scale_angular_frequency, scale, omega, amp_canonical, amp_unit_spectrum = \
        wavelet_complex(band_order_nth, time_s, time_s[-1]/2., scale_frequency_center_hz, frequency_sample_rate_hz)

    if dictionary_type == "spect":
        amp = amp_unit_spectrum
    elif dictionary_type == "unit":
        amp = 1. if np.isscalar(scale) else np.ones(scale.shape)
    else:
        amp = amp_canonical

    return amp * wavelet_gabor, xtime_shifted / frequency_sample_rate_hz, scale, omega, amp


def cwt_complex_any_scale_pow2(
        band_order_nth: float,
        sig_wf: np.ndarray,
        frequency_sample_rate_hz: float,
        cwt_type: str = "fft",
        dictionary_type: str = "norm"
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculate CWT

    :param band_order_nth: Nth order of constant Q bands
    :param sig_wf: array with input signal
    :param frequency_sample_rate_hz: sample rate in Hz
    :param cwt_type: one of "fft", or "morlet2". Default is "fft"
    :param dictionary_type: Canonical unit-norm ("norm") or unit spectrum ("spect"). Default is "norm"
    :return: frequency_cwt_hz, time_cwt_s, cwt
    """
    wavelet_points = len(sig_wf)
    time_cwt_s = np.arange(wavelet_points) / frequency_sample_rate_hz
    cycles_m = scales.cycles_from_order(scale_order=band_order_nth)

    frequency_cwt_hz = scales.log_frequency_hz_from_fft_points(
        frequency_sample_hz=frequency_sample_rate_hz,
        fft_points=len(sig_wf),
        scale_order=band_order_nth)

    cw_complex, _, _, _, amp = \
        wavelet_centered_4cwt(band_order_nth=band_order_nth,
                              duration_points=wavelet_points,
                              scale_frequency_center_hz=frequency_cwt_hz,
                              frequency_sample_rate_hz=frequency_sample_rate_hz,
                              dictionary_type=dictionary_type)

    if cwt_type == "morlet2":
        scale_atom, _ = \
            scales.scale_from_frequency_hz(scale_order=band_order_nth,
                                           frequency_sample_rate_hz=frequency_sample_rate_hz,
                                           scale_frequency_center_hz=frequency_cwt_hz)
        cwt = signal.cwt(data=sig_wf, wavelet=signal.morlet2,
                         widths=scale_atom,
                         w=cycles_m,
                         dtype=np.complex128)
        if dictionary_type == 'spect':
            # Convert to 2d matrix
            cwt *= np.tile(amplitude_convert_norm_to_spect(scale_atom=scale_atom), (wavelet_points, 1)).T

    else:
        # Convolution using the fft method
        cwt = signal.fftconvolve(np.tile(sig_wf, (len(frequency_cwt_hz), 1)),
                                 np.conj(np.fliplr(cw_complex)), mode='same', axes=-1)

    return frequency_cwt_hz, time_cwt_s, cwt



"""
This module provides benchmark synthetic functions

"""
from typing import Tuple

import numpy as np
import scipy.signal as signal

from quantum_inferno.utilities.window import get_tukey
from quantum_inferno.synth import synthetic_signals

DEFAULT_TIME_SAMPLE_INTERVAL = 1e-3
DEFAULT_TIME_DURATION = 1.0
DEFAULT_OVERSAMPLE_SCALE = 2


""" Signal conditioning  """


def signal_gate(wf: np.ndarray, t: np.ndarray, tmin: float, tmax: float, fraction_cosine: float = 0) -> np.ndarray:
    """
    Time gate and apply Tukey window, rectangular is the default

    :param wf: waveform
    :param t: timestamp array
    :param tmin: lower time limit
    :param tmax: upper time limit
    :param fraction_cosine: 0 = rectangular, 1 = Hann
    :return: waveform
    """
    index_exclude = np.logical_or(t < tmin, t > tmax)
    index_include = np.logical_and(t >= tmin, t <= tmax)
    wf[index_exclude] = 0.0
    wf[index_include] *= signal.windows.tukey(M=index_include.sum(), alpha=fraction_cosine)
    return wf


def oversample_time(time_duration: float, time_sample_interval: float, oversample_scale: float) -> np.ndarray:
    """
    Return an oversampled time by a factor oversample_scale

    :param time_duration: duration of waveform
    :param time_sample_interval: length of an interval
    :param oversample_scale: oversample waveform, then decimate by scale
    :return: oversampled timestamps
    """
    oversample_interval = time_sample_interval / oversample_scale
    number_points = int(time_duration / oversample_interval)
    time_all = np.arange(number_points) * oversample_interval
    return time_all


""" Reference Signatures, starting with the quantized Gabor chirp"""


def quantum_chirp(
    omega: float,
    order: float = 12.0,
    gamma: float = 0.0,
    gauss: bool = True,
    oversample_scale: int = DEFAULT_OVERSAMPLE_SCALE,
) -> Tuple[np.ndarray, int]:
    """
    Constructs a tone or a sweep with a gaussian window option and a duration of 2^n points

    :param omega: center frequency < pi. Resets to pi/4 if >= 1 (Nyquist).
    :param order: fractional octave band, sets the Q and duration. Default of 12 for chromatic scale.
    :param gamma: sweep index, could be positive or negative. Default of zero yields the atom.
    :param gauss: Apply the Gauss envelope, True as default. False constructs constant amplitude CW or sweep
    :param oversample_scale: oversample waveform, then decimate by scale
    :return: waveform, scale factor
    """
    if omega >= 0.8 * np.pi:
        print("Omega >= 0.8*pi (AA*Nyquist), reset to pi * 2**(-1/N")
        omega = np.pi * 2 ** (-1 / order)

    # Gabor atom specifications
    scale_multiplier = 3.0 / 4.0 * np.pi * order

    # Atom scale
    scale = scale_multiplier / omega

    # Chirp index gamma, blueshift
    chirp_scale = scale * np.sqrt(1 + gamma ** 2)

    # scale multiplier Mc
    window_support_points = 2.0 * np.pi * chirp_scale
    # scale up
    window_support_pow2 = 2 ** int((np.ceil(np.log2(window_support_points))))

    # Oversample by a factor of two:
    window_support_pow2_oversample = oversample_scale * window_support_pow2

    time0 = np.arange(window_support_pow2_oversample)
    time = time0 - time0[-1] / 2

    chirp_phase = omega * time + 0.5 * gamma * (time / chirp_scale) ** 2
    if gauss:
        chirp_wf_oversample = np.exp(-0.5 * (time / chirp_scale) ** 2 + 1j * chirp_phase)
    else:
        chirp_wf_oversample = np.exp(1j * chirp_phase)

    # Downsample to anti-alias
    chirp_wf = signal.decimate(x=np.real(chirp_wf_oversample), q=oversample_scale) + 1j * signal.decimate(
        x=np.imag(chirp_wf_oversample), q=oversample_scale
    )

    return chirp_wf, window_support_pow2


def synth_00(
    frequency_0: float = 100.0,
    frequency_1: float = 200.0,
    frequency_2: float = 400.0,
    time_start_2: float = 0.25,
    time_stop_2: float = 0.4,
    time_sample_interval: float = DEFAULT_TIME_SAMPLE_INTERVAL,
    time_duration: float = DEFAULT_TIME_DURATION,
    oversample_scale: int = DEFAULT_OVERSAMPLE_SCALE,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generate three sine waves, oversample and decimate to AA
    Always work with non-dimensionalized units (number of points, Nyquist, etc.)

    :param frequency_0: first frequency
    :param frequency_1: second frequency
    :param frequency_2: third frequency
    :param time_start_2: start time of third frequency
    :param time_stop_2: end time of third frequency
    :param time_sample_interval: length of an interval
    :param time_duration: length of waveform
    :param oversample_scale: oversample synthetic, then decimate by scale
    :return: synth waveform and timestamps
    """
    # Oversample, then decimate
    time_all = oversample_time(time_duration, time_sample_interval, oversample_scale)

    # Construct sine waves with unit amplitude [rms * sqrt(2)]
    sin_0 = np.sin(2.0 * np.pi * frequency_0 * time_all)
    signal_gate(wf=sin_0, t=time_all, tmin=0, tmax=0.5)
    sin_1 = np.sin(2.0 * np.pi * frequency_1 * time_all)
    signal_gate(wf=sin_1, t=time_all, tmin=0.5, tmax=1.0)
    sin_2 = np.sin(2.0 * np.pi * frequency_2 * time_all)
    signal_gate(wf=sin_2, t=time_all, tmin=time_start_2, tmax=time_stop_2)

    # Superpose gated sinusoids
    superpose = sin_0 + sin_1 + sin_2
    signal_gate(wf=superpose, t=time_all, tmin=0.0, tmax=1.0, fraction_cosine=0.05)

    # Decimate by same oversample scale, essentially an AA filter
    synth_wf = signal.decimate(x=superpose, q=oversample_scale)
    synth_time = np.arange(len(synth_wf)) * time_sample_interval

    return synth_wf, synth_time


def synth_01(
    a: float = 100.0,
    b: float = 20.0,
    f: float = 5.0,
    time_sample_interval: float = DEFAULT_TIME_SAMPLE_INTERVAL,
    time_duration: float = DEFAULT_TIME_DURATION,
    oversample_scale: int = DEFAULT_OVERSAMPLE_SCALE,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Example Synthetic 1

    :param a: first value
    :param b: second value
    :param f: third value
    :param time_sample_interval: length of an interval
    :param time_duration: length of waveform
    :param oversample_scale: oversample synthetic, then decimate by scale
    :return: synth waveform and timestamps
    """
    # Oversample, then decimate, essentially an AA filter
    time_all = oversample_time(time_duration, time_sample_interval, oversample_scale)
    superpose = np.cos(a * np.pi * time_all - b * np.pi * time_all * time_all) + np.cos(
        4.0 * np.pi * np.sin(np.pi * f * time_all) + np.pi * 80.0 * time_all
    )
    # Taper
    signal_gate(wf=superpose, t=time_all, tmin=0.0, tmax=1.0, fraction_cosine=0.05)
    # Decimate by same oversample scale
    synth_wf = signal.decimate(x=superpose, q=oversample_scale)
    synth_time = np.arange(len(synth_wf)) * time_sample_interval

    return synth_wf, synth_time


def synth_02(
    t1: float = 0.3,
    t2: float = 0.7,
    t3: float = 0.5,
    f1: float = 45.0,
    f2: float = 75.0,
    f3: float = 15.0,
    time_sample_interval: float = DEFAULT_TIME_SAMPLE_INTERVAL,
    time_duration: float = DEFAULT_TIME_DURATION,
    oversample_scale: int = DEFAULT_OVERSAMPLE_SCALE,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Example Synthetic 2

    :param t1: first time
    :param t2: second time
    :param t3: third time
    :param f1: first frequency
    :param f2: second frequency
    :param f3: third frequency
    :param time_sample_interval: length of an interval
    :param time_duration: length of waveform
    :param oversample_scale: oversample synthetic, then decimate by scale
    :return: synth waveform and timestamps
    """
    t = oversample_time(time_duration, time_sample_interval, oversample_scale)

    pulse1 = np.exp(-35.0 * np.pi * (t - t1) ** 2) * np.cos(np.pi * f1 * t)
    pulse2 = np.exp(-35.0 * np.pi * (t - t2) ** 2) * np.cos(np.pi * f1 * t)
    pulse3 = np.exp(-55.0 * np.pi * (t - t3) ** 2) * np.cos(np.pi * f2 * t)
    pulse4 = np.exp(-45.0 * np.pi * (t - t3) ** 2) * np.cos(np.pi * f3 * t)

    superpose = pulse1 + pulse2 + pulse3 + pulse4

    # Decimate by same oversample scale, essentially an AA filter
    synth_wf = signal.decimate(x=superpose, q=oversample_scale)
    synth_time = np.arange(len(synth_wf)) * time_sample_interval

    return synth_wf, synth_time


def synth_03(
    a: float = 30.0,
    b: float = 40.0,
    c: float = 150.0,
    time_sample_interval: float = DEFAULT_TIME_SAMPLE_INTERVAL,
    time_duration: float = DEFAULT_TIME_DURATION,
    oversample_scale: int = DEFAULT_OVERSAMPLE_SCALE,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Example Synthetic 3

    :param a: first value
    :param b: second value
    :param c: third value
    :param time_sample_interval: length of an interval
    :param time_duration: length of waveform
    :param oversample_scale: oversample synthetic, then decimate by scale
    :return: synth waveform and timestamps
    """
    # Oversample, then decimate
    time_all = oversample_time(time_duration, time_sample_interval, oversample_scale)
    superpose = np.cos(20.0 * np.pi * np.log(a * time_all + 1.0)) + np.cos(
        b * np.pi * time_all + c * np.pi * (time_all ** 2)
    )
    signal_gate(wf=superpose, t=time_all, tmin=0.0, tmax=1.0, fraction_cosine=0.05)

    # Decimate by same oversample scale, essentially an AA filter
    synth_wf = signal.decimate(x=superpose, q=oversample_scale)
    synth_time = np.arange(len(synth_wf)) * time_sample_interval

    return synth_wf, synth_time


""" Fancy test tone with time-domain specifications"""


def well_tempered_tone(
    frequency_sample_rate_hz: float = 800.0,
    frequency_center_hz: float = 60.0,
    time_duration_s: float = 10.24,
    time_fft_s: float = 0.64,
    use_fft_frequency: bool = True,
    add_noise_taper_aa: bool = False,
    output_desc: bool = False,
) -> Tuple[np.ndarray, np.ndarray, int, float, float, float]:
    """
    Return a tone of unit amplitude and fixed frequency with a constant sample rate

    :param frequency_sample_rate_hz: sample rate of frequency in Hz
    :param frequency_center_hz: center frequency in Hz
    :param time_duration_s: duration in seconds
    :param time_fft_s: length of segments. Default 1 second
    :param use_fft_frequency: use FFT frequency.  Default True
    :param add_noise_taper_aa: add noise taper.  Default False
    :param output_desc: if True, output description of waveform.  Default False
    :return: waveform, timestamps, fft duration, sample rate, center frequency, frequency resolution
    """
    # The segments determine the spectral resolution
    frequency_resolution_hz = 1.0 / time_fft_s

    # The FFT efficiency is based on powers of 2; it is always possible to pad with zeros.
    # Set the record duration, make a power of 2. Note that int rounds down
    time_duration_nd = 2 ** (int(np.log2(time_duration_s * frequency_sample_rate_hz)))
    # Set the fft duration, make a power of 2
    time_fft_nd = 2 ** (int(np.log2(time_fft_s * frequency_sample_rate_hz)))

    # Warn user if the time duration or fft duration is not a power of 2 and show the new values
    # The values are rounded down to the nearest power of 2 in the previous step.
    if time_duration_nd != time_duration_s * frequency_sample_rate_hz:
        print(
            f"Warning: The time duration {time_duration_s} s with given sample rate doesn't produce data points "
            f"that are power of two, adjusting time duration to {time_duration_nd} s"
        )
    if time_fft_nd != time_fft_s * frequency_sample_rate_hz:
        print(
            f"Warning: fft duration {time_fft_s} s with given sample rate doesn't produce data points "
            f"that are power of two, adjusting fft duration to {time_fft_nd} s"
        )

    # The fft frequencies are set by the duration of the fft
    # In this example we only need the positive frequencies
    frequency_fft_pos_hz = np.fft.rfftfreq(time_fft_nd, d=1 / frequency_sample_rate_hz)
    fft_index = np.argmin(np.abs(frequency_fft_pos_hz - frequency_center_hz))
    frequency_center_fft_hz = frequency_fft_pos_hz[fft_index]
    frequency_resolution_fft_hz = frequency_sample_rate_hz / time_fft_nd

    # Dimensionless time (samples)
    time_nd = np.arange(time_duration_nd)
    time_s = time_nd / frequency_sample_rate_hz

    # Convert to dimensionless time and frequency, which is typically used in mathematical formulas.
    # Scale by the sample rate.
    if use_fft_frequency:
        f_c = frequency_center_fft_hz / frequency_sample_rate_hz
        # Construct synthetic tone with 2^n points and max FFT amplitude at exact fft frequency
        # mic_sig = np.cos(2.0 * np.pi * frequency_center_fft * time_nd)
    else:
        # Dimensionless center frequency
        f_c = frequency_center_hz / frequency_sample_rate_hz
        # Compare to synthetic tone with 2^n points and max FFT amplitude NOT at exact fft frequency
        # It does NOT return unit amplitude (but it's close)
    mic_sig = np.cos(2.0 * np.pi * f_c * time_nd)

    if add_noise_taper_aa:
        # Add noise
        mic_sig += synthetic_signals.white_noise_fbits(sig=mic_sig, std_bit_loss=8.0)
        # Taper before AA
        mic_sig *= get_tukey(array=mic_sig, alpha=0.1)
        # Antialias (AA)
        synthetic_signals.antialias_half_nyquist(mic_sig)

    if output_desc:
        print("WELL TEMPERED TONE SYNTHETIC")
        print("Nyquist frequency:", frequency_sample_rate_hz / 2)
        print("Nominal signal frequency, hz:", frequency_center_hz)
        print("FFT signal frequency, hz:", frequency_center_fft_hz)
        print("Nominal spectral resolution, hz", frequency_resolution_hz)
        print("FFT spectral resolution, hz", frequency_resolution_fft_hz)
        print("Number of signal points:", time_duration_nd)
        print("log2(points):", np.log2(time_duration_nd))
        print("Number of FFT points:", time_fft_nd)
        print("log2(FFT points):", np.log2(time_fft_nd))

    return mic_sig, time_s, time_fft_nd, frequency_sample_rate_hz, frequency_center_fft_hz, frequency_resolution_fft_hz



"""
This module contains functions to construct quantized, standardized information packets using binary metrics.
"""

import numpy as np
import scipy.signal as signal
from quantum_inferno import scales_dyadic as scales
from typing import Tuple, Union
from quantum_inferno.utilities.rescaling import to_log2_with_epsilon


def chirp_complex(
    band_order_nth: float,
    time_s: np.ndarray,
    offset_time_s: float,
    scale_frequency_center_hz: float,
    frequency_sample_rate_hz: float,
    index_shift: float = 0,
    scale_base: float = scales.Slice.G2,
) -> Tuple[np.ndarray, np.ndarray, float, float]:
    """
    Quantum chirp for specified band_order_Nth and arbitrary time duration
    Unscaled, to be used by both Dictionary 1 and Dictionary 2

    :param band_order_nth: Nth order of constant Q bands
    :param time_s: time in seconds, duration should be greater than or equal to M/fc
    :param offset_time_s: offset time in seconds, should be between min and max of time_s
    :param scale_frequency_center_hz: center frequency fc in Hz
    :param frequency_sample_rate_hz: sample rate on Hz
    :param index_shift: Redshift = -1, Blueshift = +1, None=0
    :param scale_base: G2 or G3
    :return: waveform_complex, time_shifted_s, normal_scaling, spectrum_scaling
    """
    xtime_shifted = chirp_time(time_s, offset_time_s, frequency_sample_rate_hz)
    time_shifted_s = xtime_shifted / frequency_sample_rate_hz

    # Fundamental chirp parameters
    cycles_m, _, gamma = chirp_mqg_from_n(band_order_nth, index_shift, scale_base)
    scale_atom = chirp_scale(cycles_m, scale_frequency_center_hz, frequency_sample_rate_hz)
    p_complex = chirp_p_complex(scale_atom, gamma, index_shift)
    normal_scaling, spectrum_scaling = chirp_amplitude(scale_atom, gamma, index_shift)

    wavelet_gauss = np.exp(-p_complex * xtime_shifted ** 2)
    wavelet_gabor = wavelet_gauss * np.exp(1j * cycles_m * xtime_shifted / scale_atom)

    return wavelet_gabor, time_shifted_s, normal_scaling, spectrum_scaling


def chirp_spectrum(
    frequency_hz: np.ndarray,
    offset_time_s: float,
    band_order_nth: float,
    frequency_center_hz: float,
    frequency_sample_rate_hz: float,
    index_shift: float = 0,
    scale_base: float = scales.Slice.G2,
) -> Tuple[Union[complex, float, np.ndarray], np.ndarray]:
    """
    Spectrum of quantum wavelet for specified band_order_Nth and arbitrary time duration

    :param frequency_hz: frequency range below Nyquist
    :param offset_time_s: time of wavelet centroid
    :param band_order_nth: Nth order of constant Q bands
    :param frequency_center_hz: band center frequency in Hz
    :param frequency_sample_rate_hz: sample rate on Hz
    :param index_shift: index of shift. Default is 0.0
    :param scale_base: positive reference Base G > 1. Default is G2
    :return: Fourier transform of the Gabor atom and shifted frequency in hz
    """
    cycles_m, _, gamma = chirp_mqg_from_n(band_order_nth, index_shift, scale_base)
    scale_atom = chirp_scale(cycles_m, frequency_center_hz, frequency_sample_rate_hz)
    p_complex = chirp_p_complex(scale_atom, gamma, index_shift)

    angular_frequency_center = 2 * np.pi * frequency_center_hz / frequency_sample_rate_hz
    angular_frequency = 2 * np.pi * frequency_hz / frequency_sample_rate_hz
    offset_phase = 2 * np.pi * frequency_hz * offset_time_s
    angular_frequency_shifted = angular_frequency - angular_frequency_center
    frequency_shifted_hz = angular_frequency_shifted * frequency_sample_rate_hz / (2 * np.pi)

    spectrum_amplitude = np.sqrt(p_complex / np.abs(p_complex))
    gauss_arg = 1.0 / (4 * p_complex)
    spectrum_gauss = np.exp(-gauss_arg * (angular_frequency_shifted ** 2))
    # Phase shift from time offset
    spectrum_gabor = spectrum_amplitude * spectrum_gauss * np.exp(-1j * offset_phase)

    return spectrum_gabor, frequency_shifted_hz


def chirp_spectrum_centered(
    band_order_nth: float,
    scale_frequency_center_hz: float,
    frequency_sample_rate_hz: float,
    index_shift: float = 0,
    scale_base: float = scales.Slice.G2,
) -> Tuple[Union[complex, float, np.ndarray], np.ndarray]:
    """
    Spectrum of quantum wavelet for specified band_order_Nth and arbitrary time duration

    :param band_order_nth: Nth order of constant Q-bands
    :param scale_frequency_center_hz: band center frequency in Hz
    :param frequency_sample_rate_hz: sample rate on Hz
    :param index_shift: index of shift. Default is 0.0
    :param scale_base: positive reference Base G > 1. Default is G2
    :return: Fourier transform of the Gabor atom and shifted frequency in hz
    """
    cycles_m, _, gamma = chirp_mqg_from_n(band_order_nth, index_shift, scale_base)
    scale_atom = chirp_scale(cycles_m, scale_frequency_center_hz, frequency_sample_rate_hz)
    p_complex = chirp_p_complex(scale_atom, gamma, index_shift)
    angular_frequency_shifted = np.arange(-np.pi, np.pi, np.pi / 2 ** 7)
    frequency_shifted_hz = angular_frequency_shifted * frequency_sample_rate_hz / (2 * np.pi)

    spectrum_amplitude = np.sqrt(p_complex / np.abs(p_complex))
    spectrum_gauss = np.exp(-(angular_frequency_shifted ** 2) / (4 * p_complex))

    return spectrum_amplitude * spectrum_gauss, frequency_shifted_hz


def chirp_mqg_from_n(
    band_order_nth: float, index_shift: float = 0, scale_base: float = scales.Slice.G2
) -> Tuple[float, float, float]:
    """
    Compute the quality factor Q and multiplier M for a specified band order N.
    N is THE quantization parameter for the binary constant Q wavelet filters.

    :param band_order_nth: Band order, must be > 0.75 or reverts to N=3
    :param index_shift: index of shift. Default is 0.
    :param scale_base: positive reference Base G > 1. Default is G2
    :return: cycles M, quality factor Q, gamma
    """
    if band_order_nth < 0.7:
        band_order_nth = 3.0
        print(f"N < 0.7 specified, using N = {band_order_nth}")
    order_bandedge = scale_base ** (1.0 / 2.0 / band_order_nth)  # kN in Garces 2013
    order_scaled_bandwidth = order_bandedge - 1.0 / order_bandedge
    quality_factor_q = 1.0 / order_scaled_bandwidth  # Exact for Nth octave bands
    # Gamma is M/(2Q)
    gamma = np.sqrt(np.log(2)) * (1 - np.log(2) * (index_shift / np.pi) ** 2) ** (-0.5)
    cycles_m = 2 * quality_factor_q * gamma  # Exact, from 1/2 power points

    return cycles_m, quality_factor_q, gamma


def chirp_scale(
    cycles_m: float, scale_frequency_center_hz: Union[np.ndarray, float], frequency_sample_rate_hz: float
) -> float:
    """
    Non-dimensional scale for canonical Morlet wavelet

    :param cycles_m: number of cycles per band period
    :param scale_frequency_center_hz: scale frequency in hz
    :param frequency_sample_rate_hz: sample rate in hz
    :return: scale atom
    """
    return cycles_m * frequency_sample_rate_hz / scale_frequency_center_hz / (2.0 * np.pi)


def chirp_scale_from_order(
    band_order_nth: float,
    scale_frequency_center_hz: float,
    frequency_sample_rate_hz: float,
    index_shift: float = 0,
    scale_base: float = scales.Slice.G2,
) -> float:
    """
    Non-dimensional scale for canonical Morlet wavelet

    :param band_order_nth: Band order, must be > 0.75 or reverts to N=3
    :param scale_frequency_center_hz: scale frequency in hz
    :param frequency_sample_rate_hz: sample rate in hz
    :param index_shift: index fo shift. Default is 0.0
    :param scale_base: positive reference Base G > 1. Default is G2
    :return: scale atom
    """
    cycles_m, _, _ = chirp_mqg_from_n(band_order_nth, index_shift, scale_base)
    return chirp_scale(cycles_m, frequency_sample_rate_hz, scale_frequency_center_hz)


def chirp_uncertainty(
    scale_atom: float, frequency_sample_rate_hz: float, gamma: float, index_shift: float
) -> Tuple[float, float, float]:
    """
    Calculate the uncertainty of chirp

    :param scale_atom: from chirp_scale or chirp_scale_from_order
    :param frequency_sample_rate_hz: sample rate in hz
    :param gamma: from index_shift, M/(2Q)
    :param index_shift: index of shift
    :return: time std in seconds, frequency std in Hz, angular frequency std in Hz
    """
    time_std_s = scale_atom / np.sqrt(2) / frequency_sample_rate_hz
    angular_frequency_std = np.sqrt(1 + (index_shift * gamma) ** 2) / scale_atom / np.sqrt(2)
    angular_frequency_std_hz = frequency_sample_rate_hz * angular_frequency_std
    frequency_std_hz = angular_frequency_std_hz / 2 / np.pi

    return time_std_s, frequency_std_hz, angular_frequency_std_hz


def chirp_p_complex(scale_atom: float, gamma: float, index_shift: float) -> complex:
    """
    Fundamental chirp variable

    :param scale_atom: from chirp_scale or chirp_scale_from_order
    :param gamma: from index_shift, M/(2Q)
    :param index_shift: index of shift
    :return: p_complex
    """
    return (1 - 1j * index_shift * gamma / np.pi) / (2 * scale_atom ** 2)


def chirp_amplitude(scale_atom: float, gamma: float, index_shift: float) -> Tuple[float, float]:
    """
    Find chirp amplitude

    :param scale_atom: from chirp_scale or chirp_scale_from_order
    :param gamma: from index_shift, M/(2Q)
    :param index_shift: index of shift
    :return: normal_scaling, spectrum_scaling
    """
    p_complex = chirp_p_complex(scale_atom, gamma, index_shift)
    normal_scaling = 1 / np.pi ** 0.25 * 1 / np.sqrt(scale_atom)
    spectrum_scaling = np.sqrt(np.abs(p_complex) / np.pi)
    return normal_scaling, spectrum_scaling


def chirp_time(time_s: np.ndarray, offset_time_s: float, frequency_sample_rate_hz: float) -> np.ndarray:
    """
    Scaled time-shifted time

    :param time_s: array with time
    :param offset_time_s: offset time in seconds
    :param frequency_sample_rate_hz: sample rate in Hz
    :return: numpy array with time-shifted time
    """
    return frequency_sample_rate_hz * (time_s - offset_time_s)


def chirp_scales_from_duration(
    band_order_nth: float, sig_duration_s: float, index_shift: float = 0.0, scale_base: float = scales.Slice.G2
) -> Tuple[float, float]:
    """
    Calculate scale factor for time and frequency from chirp duration

    :param band_order_nth: Band order
    :param sig_duration_s: signal duration in seconds
    :param index_shift: index fo shift. Default is 0.0
    :param scale_base: positive reference Base G > 1. Default is G2
    :return: time in seconds and frequency in Hz scale factors
    """
    cycles_m, _, _ = chirp_mqg_from_n(band_order_nth, index_shift, scale_base)
    scale_time_s = sig_duration_s / cycles_m
    scale_frequency_hz = 1 / scale_time_s
    return scale_time_s, scale_frequency_hz


def chirp_frequency_bands(
    scale_order_input: float,
    frequency_low_input: float,
    frequency_sample_rate_input: float,
    frequency_high_input: float,
    index_shift: float = 0,
    frequency_ref: float = scales.Slice.F1HZ,
    scale_base: float = scales.Slice.G2,
) -> Tuple[float, float, float, float, np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculate frequency bands for chirp

    :param scale_order_input: Nth order specification
    :param frequency_low_input: lowest frequency of interest
    :param frequency_sample_rate_input: sample rate
    :param frequency_high_input: highest frequency of interest
    :param index_shift: index of shift
    :param frequency_ref: reference frequency
    :param scale_base: positive reference Base G > 1. Default is G2
    :return: Nth order, cycles M, quality factor Q, gamma, geometric center of frequencies, start frequency,
        end frequency
    """
    (
        order_Nth,
        scale_base,
        _,
        frequency_ref,
        _,
        frequency_center_geometric,
        frequency_start,
        frequency_end,
    ) = scales.band_frequency_low_high(
        frequency_order_input=scale_order_input,
        frequency_base_input=scale_base,
        frequency_ref_input=frequency_ref,
        frequency_low_input=frequency_low_input,
        frequency_high_input=frequency_high_input,
        frequency_sample_rate_input=frequency_sample_rate_input,
    )
    cycles_m, quality_q, gamma = chirp_mqg_from_n(order_Nth, index_shift, scale_base)

    return order_Nth, cycles_m, quality_q, gamma, frequency_center_geometric, frequency_start, frequency_end


def chirp_centered_4cwt(
    band_order_nth: float,
    sig_or_time: np.ndarray,
    scale_frequency_center_hz: float,
    frequency_sample_rate_hz: float,
    index_shift: float = 0,
    scale_base: float = scales.Slice.G2,
    dictionary_type: str = "norm",
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Gabor atoms for CWT computation centered on the duration of signal

    :param sig_or_time: time or time series, wavelet matches this duration
    :param band_order_nth: Nth order of constant Q bands
    :param scale_frequency_center_hz: center frequency fc in Hz
    :param frequency_sample_rate_hz: sample rate is Hz
    :param index_shift: index of shift
    :param scale_base: G2 or G3
    :param dictionary_type: Canonical unit-norm ("norm") or unit spectrum ("spect")
    :return: waveform_complex, time_shifted_s
    """
    duration_points = len(sig_or_time)
    time_s = np.arange(duration_points) / frequency_sample_rate_hz
    offset_time_s = time_s[-1] / 2.0

    wavelet_gabor, time_centered_s, normal_scaling, spectrum_scaling = chirp_complex(
        band_order_nth,
        time_s,
        offset_time_s,
        scale_frequency_center_hz,
        frequency_sample_rate_hz,
        index_shift,
        scale_base,
    )

    wavelet_chirp = (normal_scaling if dictionary_type == "norm" else spectrum_scaling) * wavelet_gabor

    return wavelet_chirp, time_centered_s


def cwt_chirp_complex(
    band_order_nth: float,
    sig_wf: np.ndarray,
    frequency_low_hz: float,
    frequency_sample_rate_hz: float,
    frequency_high_hz: float = scales.Slice.F0HZ,
    cwt_type: str = "fft",
    index_shift: float = 0,
    frequency_ref: float = scales.Slice.F1HZ,
    scale_base: float = scales.Slice.G2,
    dictionary_type: str = "norm",
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculate CWT for chirp

    :param band_order_nth: Nth order of constant Q bands
    :param sig_wf: array with input signal
    :param frequency_low_hz: lowest frequency in Hz
    :param frequency_sample_rate_hz: sample rate in Hz
    :param frequency_high_hz: highest frequency in Hz
    :param cwt_type: one of "conv", "fft", or "morlet2". Default is "fft"
           Address ghost folding in "fft", compared to "conv"
    :param index_shift: index of shift. Default is 0.0
    :param frequency_ref: reference frequency in Hz. Default is F1HZ
    :param scale_base: G2 or G3. Default is G2
    :param dictionary_type: Canonical unit-norm ("norm") or unit spectrum ("spect"). Default is "norm"
    :return: cwt, cwt_bits, time_s, frequency_cwt_hz
    """
    wavelet_points = len(sig_wf)
    time_s = np.arange(wavelet_points) / frequency_sample_rate_hz

    if cwt_type == "morlet2":
        index_shift = 0

    # Planck frequency is absolute upper limit
    if frequency_high_hz > frequency_sample_rate_hz / 2.0:
        frequency_high_hz = frequency_sample_rate_hz / 2.0

    (
        order_Nth,
        cycles_M,
        _,
        _,
        frequency_cwt_hz_flipped,
        frequency_start_flipped,
        frequency_end_flipped,
    ) = chirp_frequency_bands(
        scale_order_input=band_order_nth,
        frequency_low_input=frequency_low_hz,
        frequency_sample_rate_input=frequency_sample_rate_hz,
        frequency_high_input=frequency_high_hz,
        index_shift=index_shift,
        frequency_ref=frequency_ref,
        scale_base=scale_base,
    )

    scale_points = len(frequency_cwt_hz_flipped)

    if cwt_type == "morlet2":
        scale_atom = chirp_scale(cycles_M, frequency_cwt_hz_flipped, frequency_sample_rate_hz)
        cwt_flipped = signal.cwt(
            data=sig_wf, wavelet=signal.morlet2, widths=scale_atom, w=cycles_M, dtype=np.complex128
        )
    elif cwt_type == "fft":
        sig_fft = np.fft.fft(sig_wf)
        cwt_flipped = np.empty((scale_points, wavelet_points), dtype=np.complex128)
        for ii in range(scale_points):
            atom, _ = chirp_centered_4cwt(
                band_order_nth=order_Nth,
                sig_or_time=sig_wf,
                scale_frequency_center_hz=frequency_cwt_hz_flipped[ii],
                frequency_sample_rate_hz=frequency_sample_rate_hz,
                index_shift=index_shift,
                scale_base=scale_base,
                dictionary_type=dictionary_type,
            )
            atom_fft = np.fft.fft(atom)
            cwt_raw = np.fft.ifft(sig_fft * np.conj(atom_fft))
            cwt_flipped[ii, :] = np.append(cwt_raw[wavelet_points // 2:], cwt_raw[0: wavelet_points // 2])

    elif cwt_type == "conv":
        cwt_flipped = np.empty((scale_points, wavelet_points), dtype=np.complex128)
        for ii in range(scale_points):
            atom, _ = chirp_centered_4cwt(
                band_order_nth=order_Nth,
                sig_or_time=sig_wf,
                scale_frequency_center_hz=frequency_cwt_hz_flipped[ii],
                frequency_sample_rate_hz=frequency_sample_rate_hz,
                index_shift=index_shift,
                scale_base=scale_base,
                dictionary_type=dictionary_type,
            )
            cwt_flipped[ii, :] = signal.convolve(sig_wf, np.conj(atom)[::-1], mode="same")
    else:
        raise ValueError(f"Incorrect cwt_type: {cwt_type} specified in cwt_chirp_complex")

    # Time scales are increasing, which is the opposite of what is expected for the frequency. Flip.
    frequency_cwt_hz = np.flip(frequency_cwt_hz_flipped)
    cwt = np.flipud(cwt_flipped)
    cwt_bits = to_log2_with_epsilon(cwt)

    return cwt, cwt_bits, time_s, frequency_cwt_hz


def cwt_chirp_from_sig(
    sig_wf: np.ndarray,
    frequency_sample_rate_hz: float,
    band_order_nth: float = 3,
    cwt_type: str = "fft",
    index_shift: float = 0,
    frequency_ref: float = scales.Slice.F1HZ,
    scale_base: float = scales.Slice.G2,
    dictionary_type: str = "norm",
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculate CWT for chirp

    :param sig_wf: array with input signal
    :param frequency_sample_rate_hz: sample rate in Hz
    :param band_order_nth: Nth order of constant Q bands
    :param cwt_type: one of "conv", "fft", or "morlet2". Default is "fft"
    :param index_shift: index of shift. Default is 0.0
    :param frequency_ref: reference frequency in Hz. Default is F1HZ
    :param scale_base: G2 or G3. Default is G2
    :param dictionary_type: Canonical unit-norm ("norm") or unit spectrum ("spect"). Default is "norm"
    :return: cwt, cwt_bits, time_s, frequency_cwt_hz
    """
    duration_s = len(sig_wf) / frequency_sample_rate_hz
    _, min_frequency_hz = chirp_scales_from_duration(
        band_order_nth=band_order_nth, sig_duration_s=duration_s, index_shift=index_shift, scale_base=scale_base
    )

    return cwt_chirp_complex(
        band_order_nth=band_order_nth,
        sig_wf=sig_wf,
        frequency_low_hz=min_frequency_hz,
        frequency_sample_rate_hz=frequency_sample_rate_hz,
        frequency_high_hz=frequency_sample_rate_hz / 2.0,
        cwt_type=cwt_type,
        index_shift=index_shift,
        frequency_ref=frequency_ref,
        scale_base=scale_base,
        dictionary_type=dictionary_type,
    )
